@@ -13,9 +13,7 @@ use base64::prelude::{Engine, BASE64_STANDARD};
 use pki_types::{CertificateDer, PrivateKeyDer, ServerName, UnixTime};
 use rustls::client::danger::{HandshakeSignatureValid, ServerCertVerified, ServerCertVerifier};
 use rustls::client::{ClientConfig, ClientConnection, Resumption, WebPkiServerVerifier};
-#[cfg(all(not(feature = "ring"), feature = "aws_lc_rs"))]
-use rustls::crypto::aws_lc_rs as provider;
-#[cfg(feature = "ring")]
+#[cfg(all(feature = "ring", not(feature = "aws_lc_rs")))]
 use rustls::crypto::ring as provider;
 use rustls::crypto::{CryptoProvider, SupportedKxGroup};
 use rustls::internal::msgs::codec::Codec;
@@ -23,10 +21,19 @@ use rustls::internal::msgs::persist::ServerSessionValue;
 use rustls::server::danger::{ClientCertVerified, ClientCertVerifier};
 use rustls::server::{ClientHello, ServerConfig, ServerConnection, WebPkiClientVerifier};
 use rustls::{
-    client, server, sign, version, AlertDescription, CertificateError, Connection,
-    DigitallySignedStruct, DistinguishedName, Error, InvalidMessage, NamedGroup, PeerIncompatible,
-    PeerMisbehaved, ProtocolVersion, RootCertStore, Side, SignatureAlgorithm, SignatureScheme,
-    SupportedProtocolVersion,
+    client, compress, server, sign, version, AlertDescription, CertificateCompressionAlgorithm,
+    CertificateError, Connection, DigitallySignedStruct, DistinguishedName, Error, HandshakeKind,
+    InvalidMessage, NamedGroup, PeerIncompatible, PeerMisbehaved, ProtocolVersion, RootCertStore,
+    Side, SignatureAlgorithm, SignatureScheme, SupportedProtocolVersion,
+};
+#[cfg(feature = "aws_lc_rs")]
+use rustls::{
+    client::{EchConfig, EchGreaseConfig, EchMode, EchStatus},
+    crypto::aws_lc_rs as provider,
+    crypto::aws_lc_rs::hpke,
+    crypto::hpke::{Hpke, HpkePublicKey},
+    internal::msgs::codec::Reader,
+    internal::msgs::handshake::EchConfigPayload,
 };
 
 static BOGO_NACK: i32 = 89;
@@ -84,6 +91,17 @@ struct Options {
     resumption_delay: u32,
     queue_early_data_after_received_messages: Vec<usize>,
     require_ems: bool,
+    expect_handshake_kind: Option<Vec<HandshakeKind>>,
+    expect_handshake_kind_resumed: Option<Vec<HandshakeKind>>,
+    install_cert_compression_algs: CompressionAlgs,
+    provider: CryptoProvider,
+    ech_config_list: Option<pki_types::EchConfigListBytes<'static>>,
+    expect_ech_accept: bool,
+    expect_ech_retry_configs: Option<pki_types::EchConfigListBytes<'static>>,
+    on_resume_ech_config_list: Option<pki_types::EchConfigListBytes<'static>>,
+    on_resume_expect_ech_accept: bool,
+    on_initial_expect_ech_accept: bool,
+    enable_ech_grease: bool,
 }
 
 impl Options {
@@ -134,6 +152,17 @@ impl Options {
             resumption_delay: 0,
             queue_early_data_after_received_messages: vec![],
             require_ems: false,
+            expect_handshake_kind: None,
+            expect_handshake_kind_resumed: Some(vec![HandshakeKind::Resumed]),
+            install_cert_compression_algs: CompressionAlgs::None,
+            provider: default_provider(),
+            ech_config_list: None,
+            expect_ech_accept: false,
+            expect_ech_retry_configs: None,
+            on_resume_ech_config_list: None,
+            on_resume_expect_ech_accept: false,
+            on_initial_expect_ech_accept: false,
+            enable_ech_grease: false,
         }
     }
 
@@ -165,6 +194,16 @@ impl Options {
     }
 }
 
+fn default_provider() -> CryptoProvider {
+    // ensure all suites and kx groups are included (even in fips builds)
+    // as non-fips test cases require them
+    CryptoProvider {
+        kx_groups: provider::ALL_KX_GROUPS.to_vec(),
+        cipher_suites: provider::ALL_CIPHER_SUITES.to_vec(),
+        ..provider::default_provider()
+    }
+}
+
 fn load_cert(filename: &str) -> Vec<CertificateDer<'static>> {
     let certfile = fs::File::open(filename).expect("cannot open certificate file");
     let mut reader = BufReader::new(certfile);
@@ -516,13 +555,13 @@ fn make_server_cfg(opts: &Options) -> Arc<ServerConfig> {
             .map(|curveid| lookup_kx_group(*curveid))
             .collect()
     } else {
-        provider::ALL_KX_GROUPS.to_vec()
+        opts.provider.kx_groups.clone()
     };
 
     let mut cfg = ServerConfig::builder_with_provider(
         CryptoProvider {
             kx_groups,
-            ..provider::default_provider()
+            ..opts.provider.clone()
         }
         .into(),
     )
@@ -569,6 +608,20 @@ fn make_server_cfg(opts: &Options) -> Arc<ServerConfig> {
         cfg.send_half_rtt_data = true;
     }
 
+    match opts.install_cert_compression_algs {
+        CompressionAlgs::All => {
+            cfg.cert_compressors = vec![&ExpandingAlgorithm, &ShrinkingAlgorithm, &RandomAlgorithm];
+            cfg.cert_decompressors =
+                vec![&ExpandingAlgorithm, &ShrinkingAlgorithm, &RandomAlgorithm];
+        }
+        CompressionAlgs::One(ShrinkingAlgorithm::ALGORITHM) => {
+            cfg.cert_compressors = vec![&ShrinkingAlgorithm];
+            cfg.cert_decompressors = vec![&ShrinkingAlgorithm];
+        }
+        CompressionAlgs::None => {}
+        _ => unimplemented!(),
+    }
+
     Arc::new(cfg)
 }
 
@@ -649,20 +702,46 @@ fn make_client_cfg(opts: &Options) -> Arc<ClientConfig> {
             .map(|curveid| lookup_kx_group(*curveid))
             .collect()
     } else {
-        provider::ALL_KX_GROUPS.to_vec()
+        opts.provider.kx_groups.clone()
     };
 
     let cfg = ClientConfig::builder_with_provider(
         CryptoProvider {
             kx_groups,
-            ..provider::default_provider()
+            ..opts.provider.clone()
         }
         .into(),
-    )
-    .with_protocol_versions(&opts.supported_versions())
-    .expect("inconsistent settings")
-    .dangerous()
-    .with_custom_certificate_verifier(Arc::new(DummyServerAuth::new()));
+    );
+
+    #[cfg(feature = "aws_lc_rs")]
+    let cfg = if let Some(ech_config_list) = &opts.ech_config_list {
+        let ech_mode: EchMode = EchConfig::new(ech_config_list.clone(), hpke::ALL_SUPPORTED_SUITES)
+            .unwrap_or_else(|_| quit(":INVALID_ECH_CONFIG_LIST:"))
+            .into();
+
+        cfg.with_ech(ech_mode)
+            .expect("invalid ECH config")
+    } else if opts.enable_ech_grease {
+        let ech_mode = EchMode::Grease(EchGreaseConfig::new(
+            GREASE_HPKE_SUITE,
+            HpkePublicKey(GREASE_25519_PUBKEY.to_vec()),
+        ));
+
+        cfg.with_ech(ech_mode)
+            .expect("invalid GREASE ECH config")
+    } else {
+        cfg.with_protocol_versions(&opts.supported_versions())
+            .expect("inconsistent settings")
+    };
+
+    #[cfg(not(feature = "aws_lc_rs"))]
+    let cfg = cfg
+        .with_protocol_versions(&opts.supported_versions())
+        .expect("inconsistent settings");
+
+    let cfg = cfg
+        .dangerous()
+        .with_custom_certificate_verifier(Arc::new(DummyServerAuth::new()));
 
     let mut cfg = if !opts.cert_file.is_empty() && !opts.key_file.is_empty() {
         let cert = load_cert(&opts.cert_file);
@@ -698,6 +777,20 @@ fn make_client_cfg(opts: &Options) -> Arc<ClientConfig> {
         cfg.enable_early_data = true;
     }
 
+    match opts.install_cert_compression_algs {
+        CompressionAlgs::All => {
+            cfg.cert_decompressors =
+                vec![&ExpandingAlgorithm, &ShrinkingAlgorithm, &RandomAlgorithm];
+            cfg.cert_compressors = vec![&ExpandingAlgorithm, &ShrinkingAlgorithm, &RandomAlgorithm];
+        }
+        CompressionAlgs::One(ShrinkingAlgorithm::ALGORITHM) => {
+            cfg.cert_decompressors = vec![&ShrinkingAlgorithm];
+            cfg.cert_compressors = vec![&ShrinkingAlgorithm];
+        }
+        CompressionAlgs::None => {}
+        _ => unimplemented!(),
+    }
+
     Arc::new(cfg)
 }
 
@@ -711,7 +804,7 @@ fn quit_err(why: &str) -> ! {
     process::exit(1)
 }
 
-fn handle_err(err: Error) -> ! {
+fn handle_err(opts: &Options, err: Error) -> ! {
     println!("TLS error: {:?}", err);
     thread::sleep(time::Duration::from_millis(100));
 
@@ -749,12 +842,31 @@ fn handle_err(err: Error) -> ! {
         | Error::InvalidMessage(InvalidMessage::InvalidEmptyPayload)
         | Error::InvalidMessage(InvalidMessage::UnknownProtocolVersion)
         | Error::InvalidMessage(InvalidMessage::MessageTooLarge) => quit(":GARBAGE:"),
+        Error::InvalidMessage(InvalidMessage::MessageTooShort)
+            if opts.enable_ech_grease || opts.ech_config_list.is_some() =>
+        {
+            quit(":ERROR_PARSING_EXTENSION:")
+        }
         Error::InvalidMessage(InvalidMessage::UnexpectedMessage(_)) => quit(":GARBAGE:"),
+        Error::DecryptError if opts.ech_config_list.is_some() => {
+            quit(":INCONSISTENT_ECH_NEGOTIATION:")
+        }
         Error::DecryptError => quit(":DECRYPTION_FAILED_OR_BAD_RECORD_MAC:"),
         Error::NoApplicationProtocol => quit(":NO_APPLICATION_PROTOCOL:"),
         Error::PeerIncompatible(
             PeerIncompatible::ServerSentHelloRetryRequestWithUnknownExtension,
         ) => quit(":UNEXPECTED_EXTENSION:"),
+        Error::PeerIncompatible(PeerIncompatible::ServerRejectedEncryptedClientHello(
+            _retry_configs,
+        )) => {
+            #[cfg(feature = "aws_lc_rs")]
+            if let Some(expected_configs) = &opts.expect_ech_retry_configs {
+                let expected_configs =
+                    Vec::<EchConfigPayload>::read(&mut Reader::init(expected_configs)).unwrap();
+                assert_eq!(_retry_configs, Some(expected_configs));
+            }
+            quit(":ECH_REJECTED:")
+        }
         Error::PeerIncompatible(_) => quit(":INCOMPATIBLE:"),
         Error::PeerMisbehaved(PeerMisbehaved::MissingPskModesExtension) => {
             quit(":MISSING_EXTENSION:")
@@ -766,6 +878,40 @@ fn handle_err(err: Error) -> ! {
         | Error::PeerMisbehaved(PeerMisbehaved::SignedKxWithWrongAlgorithm) => {
             quit(":WRONG_SIGNATURE_TYPE:")
         }
+        Error::PeerMisbehaved(PeerMisbehaved::SelectedUnofferedCertCompression) => {
+            quit(":UNKNOWN_CERT_COMPRESSION_ALG:")
+        }
+        Error::PeerMisbehaved(PeerMisbehaved::InvalidCertCompression) => {
+            quit(":CERT_DECOMPRESSION_FAILED:")
+        }
+        Error::PeerMisbehaved(PeerMisbehaved::OfferedDuplicateCertificateCompressions) => {
+            quit(":ERROR_PARSING_EXTENSION:")
+        }
+        Error::PeerMisbehaved(PeerMisbehaved::SelectedUnofferedCipherSuite) => {
+            quit(":WRONG_CIPHER_RETURNED:")
+        }
+        Error::PeerMisbehaved(PeerMisbehaved::TooManyWarningAlertsReceived) => {
+            quit(":TOO_MANY_WARNING_ALERTS:")
+        }
+        Error::PeerMisbehaved(PeerMisbehaved::TooManyKeyUpdateRequests) => {
+            quit(":TOO_MANY_KEY_UPDATES:")
+        }
+        Error::PeerMisbehaved(PeerMisbehaved::TooManyEmptyFragments) => {
+            quit(":TOO_MANY_EMPTY_FRAGMENTS:")
+        }
+        Error::PeerMisbehaved(PeerMisbehaved::IllegalHelloRetryRequestWithInvalidEch)
+        | Error::PeerMisbehaved(PeerMisbehaved::UnsolicitedEchExtension) => {
+            quit(":UNEXPECTED_EXTENSION:")
+        }
+        // The TLS-ECH-Client-UnsolicitedInnerServerNameAck test is expected to fail with
+        // :UNEXPECTED_EXTENSION: when we receive an unsolicited inner hello SNI extension.
+        // We treat this the same as any unexpected enc'd ext and return :PEER_MISBEHAVIOUR:.
+        // Convert to the expected if this error occurs when we're configured w/ ECH.
+        Error::PeerMisbehaved(PeerMisbehaved::UnsolicitedEncryptedExtension)
+            if opts.ech_config_list.is_some() =>
+        {
+            quit(":UNEXPECTED_EXTENSION:")
+        }
         Error::PeerMisbehaved(_) => quit(":PEER_MISBEHAVIOUR:"),
         Error::NoCertificatesPresented => quit(":NO_CERTS:"),
         Error::AlertReceived(AlertDescription::UnexpectedMessage) => quit(":BAD_ALERT:"),
@@ -808,14 +954,14 @@ fn server(conn: &mut Connection) -> &mut ServerConnection {
 
 const MAX_MESSAGE_SIZE: usize = 0xffff + 5;
 
-fn after_read(sess: &mut Connection, conn: &mut net::TcpStream) {
+fn after_read(opts: &Options, sess: &mut Connection, conn: &mut net::TcpStream) {
     if let Err(err) = sess.process_new_packets() {
         flush(sess, conn); /* send any alerts before exiting */
-        handle_err(err);
+        handle_err(opts, err);
     }
 }
 
-fn read_n_bytes(sess: &mut Connection, conn: &mut net::TcpStream, n: usize) {
+fn read_n_bytes(opts: &Options, sess: &mut Connection, conn: &mut net::TcpStream, n: usize) {
     let mut bytes = [0u8; MAX_MESSAGE_SIZE];
     match conn.read(&mut bytes[..n]) {
         Ok(count) => {
@@ -827,17 +973,17 @@ fn read_n_bytes(sess: &mut Connection, conn: &mut net::TcpStream, n: usize) {
         Err(err) => panic!("invalid read: {}", err),
     };
 
-    after_read(sess, conn);
+    after_read(opts, sess, conn);
 }
 
-fn read_all_bytes(sess: &mut Connection, conn: &mut net::TcpStream) {
+fn read_all_bytes(opts: &Options, sess: &mut Connection, conn: &mut net::TcpStream) {
     match sess.read_tls(conn) {
         Ok(_) => {}
         Err(ref err) if err.kind() == io::ErrorKind::ConnectionReset => {}
         Err(err) => panic!("invalid read: {}", err),
     };
 
-    after_read(sess, conn);
+    after_read(opts, sess, conn);
 }
 
 fn exec(opts: &Options, mut sess: Connection, count: usize) {
@@ -863,7 +1009,7 @@ fn exec(opts: &Options, mut sess: Connection, count: usize) {
             {
                 flush(&mut sess, &mut conn);
                 for message_size_estimate in &opts.queue_early_data_after_received_messages {
-                    read_n_bytes(&mut sess, &mut conn, *message_size_estimate);
+                    read_n_bytes(opts, &mut sess, &mut conn, *message_size_estimate);
                 }
                 println!("now ready for early data");
             }
@@ -889,7 +1035,7 @@ fn exec(opts: &Options, mut sess: Connection, count: usize) {
         }
 
         if sess.wants_read() {
-            read_all_bytes(&mut sess, &mut conn);
+            read_all_bytes(opts, &mut sess, &mut conn);
         }
 
         if opts.side == Side::Server && opts.enable_early_data {
@@ -967,6 +1113,30 @@ fn exec(opts: &Options, mut sess: Connection, count: usize) {
             }
         }
 
+        if opts.expect_handshake_kind.is_some() && !sess.is_handshaking() {
+            let expected_options = opts
+                .expect_handshake_kind
+                .as_ref()
+                .unwrap();
+            let actual = sess.handshake_kind().unwrap();
+            assert!(
+                expected_options.contains(&actual),
+                "wanted to see {expected_options:?} but got {actual:?}"
+            );
+        }
+
+        #[cfg(feature = "aws_lc_rs")]
+        {
+            let ech_accept_required =
+                (count == 0 && opts.on_initial_expect_ech_accept) || opts.expect_ech_accept;
+            if ech_accept_required
+                && !sess.is_handshaking()
+                && client(&mut sess).ech_status() != EchStatus::Accepted
+            {
+                quit_err("ECH was not accepted, but we expect the opposite");
+            }
+        }
+
         let mut buf = [0u8; 1024];
         let len = match sess
             .reader()
@@ -1131,8 +1301,6 @@ pub fn main() {
             "-expect-secure-renegotiation" |
             "-expect-no-session-id" |
             "-enable-ed25519" |
-            "-expect-hrr" |
-            "-expect-no-hrr" |
             "-on-resume-expect-no-offer-early-data" |
             "-key-update" | //< we could implement an API for this
             "-expect-tls13-downgrade" |
@@ -1141,6 +1309,18 @@ pub fn main() {
                 println!("not checking {}; NYI", arg);
             }
 
+            "-expect-hrr" => {
+                opts.expect_handshake_kind = Some(vec![HandshakeKind::FullWithHelloRetryRequest]);
+            }
+            "-expect-no-hrr" => {
+                opts.expect_handshake_kind = Some(vec![HandshakeKind::Full]);
+            }
+            "-expect-session-miss" => {
+                opts.expect_handshake_kind_resumed = Some(vec![
+                    HandshakeKind::Full,
+                    HandshakeKind::FullWithHelloRetryRequest
+                ]);
+            }
             "-export-keying-material" => {
                 opts.export_keying_material = args.remove(0).parse::<usize>().unwrap();
             }
@@ -1265,6 +1445,53 @@ pub fn main() {
             "-expect-extended-master-secret" => {
                 opts.require_ems = true;
             }
+            "-install-cert-compression-algs" => {
+                opts.install_cert_compression_algs = CompressionAlgs::All;
+            }
+            "-install-one-cert-compression-alg" => {
+                opts.install_cert_compression_algs = CompressionAlgs::One(args.remove(0).parse::<u16>().unwrap());
+            }
+            #[cfg(feature = "fips")]
+            "-fips-202205" => {
+                opts.provider = rustls::crypto::default_fips_provider();
+            }
+            #[cfg(not(feature = "fips"))]
+            "-fips-202205" => {
+                println!("Not a FIPS build");
+                process::exit(BOGO_NACK);
+            }
+            "-ech-config-list" => {
+                opts.ech_config_list = Some(BASE64_STANDARD.decode(args.remove(0).as_bytes())
+                    .expect("invalid ECH config base64").into());
+            }
+            "-expect-ech-accept" => {
+                opts.expect_ech_accept = true;
+            }
+            "-expect-ech-retry-configs" => {
+                opts.expect_ech_retry_configs = Some(BASE64_STANDARD.decode(args.remove(0).as_bytes())
+                    .expect("invalid ECH config base64").into());
+            }
+            "-on-resume-ech-config-list" => {
+                opts.on_resume_ech_config_list = Some(BASE64_STANDARD.decode(args.remove(0).as_bytes())
+                    .expect("invalid on resume ECH config base64").into());
+            }
+            "-on-resume-expect-ech-accept" => {
+                opts.on_resume_expect_ech_accept = true;
+            }
+            "-expect-no-ech-retry-configs" => {
+                opts.expect_ech_retry_configs = None;
+            }
+            "-on-initial-expect-ech-accept" => {
+                opts.on_initial_expect_ech_accept = true;
+            }
+            "-on-retry-expect-ech-retry-configs" => {
+                // Note: we treat this the same as -expect-ech-retry-configs
+                opts.expect_ech_retry_configs = Some(BASE64_STANDARD.decode(args.remove(0).as_bytes())
+                    .expect("invalid retry ECH config base64").into());
+            }
+            "-enable-ech-grease" => {
+                opts.enable_ech_grease = true;
+            }
 
             // defaults:
             "-enable-all-curves" |
@@ -1276,9 +1503,9 @@ pub fn main() {
             "-ipv6" |
             "-decline-alpn" |
             "-expect-no-session" |
-            "-expect-session-miss" |
             "-expect-ticket-renewal" |
             "-enable-ocsp-stapling" |
+            "-forbid-renegotiation-after-handshake" |
             // internal openssl details:
             "-async" |
             "-implicit-handshake" |
@@ -1330,12 +1557,12 @@ pub fn main() {
             "-no-rsa-pss-rsae-certs" |
             "-ignore-tls13-downgrade" |
             "-allow-hint-mismatch" |
-            "-fips-202205" |
             "-wpa-202304" |
             "-srtp-profiles" |
             "-permute-extensions" |
             "-signed-cert-timestamps" |
-            "-on-initial-expect-peer-cert-file" => {
+            "-on-initial-expect-peer-cert-file" |
+            "-use-custom-verify-callback" => {
                 println!("NYI option {:?}", arg);
                 process::exit(BOGO_NACK);
             }
@@ -1349,7 +1576,7 @@ pub fn main() {
 
     println!("opts {:?}", opts);
 
-    let (client_cfg, mut server_cfg) = match opts.side {
+    let (mut client_cfg, mut server_cfg) = match opts.side {
         Side::Client => (Some(make_client_cfg(&opts)), None),
         Side::Server => (None, Some(make_server_cfg(&opts))),
     };
@@ -1388,5 +1615,157 @@ pub fn main() {
             opts.tickets = false;
             server_cfg = Some(make_server_cfg(&opts));
         }
+        if opts.on_resume_ech_config_list.is_some() {
+            opts.ech_config_list
+                .clone_from(&opts.on_resume_ech_config_list);
+            opts.expect_ech_accept = opts.on_resume_expect_ech_accept;
+            client_cfg = Some(make_client_cfg(&opts));
+        }
+        opts.expect_handshake_kind
+            .clone_from(&opts.expect_handshake_kind_resumed);
     }
 }
+
+#[derive(Debug, PartialEq)]
+enum CompressionAlgs {
+    None,
+    All,
+    One(u16),
+}
+
+#[derive(Debug)]
+struct ShrinkingAlgorithm;
+
+impl ShrinkingAlgorithm {
+    const ALGORITHM: u16 = 0xff01;
+}
+
+impl compress::CertDecompressor for ShrinkingAlgorithm {
+    fn algorithm(&self) -> CertificateCompressionAlgorithm {
+        CertificateCompressionAlgorithm::Unknown(Self::ALGORITHM)
+    }
+
+    fn decompress(
+        &self,
+        input: &[u8],
+        output: &mut [u8],
+    ) -> Result<(), compress::DecompressionFailed> {
+        if output.len() != input.len() + 2 {
+            return Err(compress::DecompressionFailed);
+        }
+        output[..2].copy_from_slice(&[0, 0]);
+        output[2..].copy_from_slice(input);
+        Ok(())
+    }
+}
+
+impl compress::CertCompressor for ShrinkingAlgorithm {
+    fn algorithm(&self) -> CertificateCompressionAlgorithm {
+        CertificateCompressionAlgorithm::Unknown(Self::ALGORITHM)
+    }
+
+    fn compress(
+        &self,
+        mut input: Vec<u8>,
+        _: compress::CompressionLevel,
+    ) -> Result<Vec<u8>, compress::CompressionFailed> {
+        assert_eq!(input[..2], [0, 0]);
+        input.drain(0..2);
+        Ok(input)
+    }
+}
+
+#[derive(Debug)]
+struct ExpandingAlgorithm;
+
+impl compress::CertDecompressor for ExpandingAlgorithm {
+    fn algorithm(&self) -> CertificateCompressionAlgorithm {
+        CertificateCompressionAlgorithm::Unknown(0xff02)
+    }
+
+    fn decompress(
+        &self,
+        input: &[u8],
+        output: &mut [u8],
+    ) -> Result<(), compress::DecompressionFailed> {
+        if output.len() + 4 != input.len() {
+            return Err(compress::DecompressionFailed);
+        }
+        if input[..4] != [1, 2, 3, 4] {
+            return Err(compress::DecompressionFailed);
+        }
+        output.copy_from_slice(&input[4..]);
+        Ok(())
+    }
+}
+
+impl compress::CertCompressor for ExpandingAlgorithm {
+    fn algorithm(&self) -> CertificateCompressionAlgorithm {
+        CertificateCompressionAlgorithm::Unknown(0xff02)
+    }
+
+    fn compress(
+        &self,
+        mut input: Vec<u8>,
+        _: compress::CompressionLevel,
+    ) -> Result<Vec<u8>, compress::CompressionFailed> {
+        input.insert(0, 1);
+        input.insert(1, 2);
+        input.insert(2, 3);
+        input.insert(3, 4);
+        Ok(input)
+    }
+}
+
+#[derive(Debug)]
+struct RandomAlgorithm;
+
+impl compress::CertDecompressor for RandomAlgorithm {
+    fn algorithm(&self) -> CertificateCompressionAlgorithm {
+        CertificateCompressionAlgorithm::Unknown(0xff03)
+    }
+
+    fn decompress(
+        &self,
+        input: &[u8],
+        output: &mut [u8],
+    ) -> Result<(), compress::DecompressionFailed> {
+        if output.len() + 1 != input.len() {
+            return Err(compress::DecompressionFailed);
+        }
+        output.copy_from_slice(&input[1..]);
+        Ok(())
+    }
+}
+
+impl compress::CertCompressor for RandomAlgorithm {
+    fn algorithm(&self) -> CertificateCompressionAlgorithm {
+        CertificateCompressionAlgorithm::Unknown(0xff03)
+    }
+
+    fn compress(
+        &self,
+        mut input: Vec<u8>,
+        _: compress::CompressionLevel,
+    ) -> Result<Vec<u8>, compress::CompressionFailed> {
+        let random_byte = {
+            let mut bytes = [0];
+            provider::default_provider()
+                .secure_random
+                .fill(&mut bytes)
+                .unwrap();
+            bytes[0]
+        };
+        input.insert(0, random_byte);
+        Ok(input)
+    }
+}
+
+#[cfg(feature = "aws_lc_rs")]
+static GREASE_HPKE_SUITE: &dyn Hpke = hpke::DH_KEM_X25519_HKDF_SHA256_AES_128;
+
+#[cfg(feature = "aws_lc_rs")]
+const GREASE_25519_PUBKEY: &[u8] = &[
+    0x67, 0x35, 0xCA, 0x50, 0x21, 0xFC, 0x4F, 0xE6, 0x29, 0x3B, 0x31, 0x2C, 0xB5, 0xE0, 0x97, 0xD8,
+    0xD0, 0x58, 0x97, 0xCF, 0x5C, 0x15, 0x12, 0x79, 0x4B, 0xEF, 0x1D, 0x98, 0x52, 0x74, 0xDC, 0x5E,
+];
