@@ -13,7 +13,7 @@ use super::hs::ClientContext;
 use crate::check::{inappropriate_handshake_message, inappropriate_message};
 use crate::client::common::{ClientAuthDetails, ServerCertDetails};
 use crate::client::{hs, ClientConfig};
-use crate::common_state::{CommonState, Side, State};
+use crate::common_state::{CommonState, HandshakeKind, Side, State};
 use crate::conn::ConnectionRandoms;
 use crate::crypto::KeyExchangeAlgorithm;
 use crate::enums::{AlertDescription, ContentType, HandshakeType, ProtocolVersion};
@@ -138,6 +138,7 @@ mod server_hello {
                             .clone()
                             .into_owned(),
                     );
+                    cx.common.handshake_kind = Some(HandshakeKind::Resumed);
                     let cert_verified = verify::ServerCertVerified::assertion();
                     let sig_verified = verify::HandshakeSignatureValid::assertion();
 
@@ -172,6 +173,7 @@ mod server_hello {
                 }
             }
 
+            cx.common.handshake_kind = Some(HandshakeKind::Full);
             Ok(Box::new(ExpectCertificate {
                 config: self.config,
                 resuming_session: None,
@@ -739,6 +741,7 @@ impl State<ClientConnectionData> for ExpectCertificateRequest<'_> {
         // is entirely duplicated in certreq.sigschemes.
 
         const NO_CONTEXT: Option<Vec<u8>> = None; // TLS 1.2 doesn't use a context.
+        let no_compression = None; // or compression
         let client_auth = ClientAuthDetails::resolve(
             self.config
                 .client_auth_cert_resolver
@@ -746,6 +749,7 @@ impl State<ClientConnectionData> for ExpectCertificateRequest<'_> {
             Some(&certreq.canames),
             &certreq.sigschemes,
             NO_CONTEXT,
+            no_compression,
         );
 
         Ok(Box::new(ExpectServerDone {
@@ -880,11 +884,12 @@ impl State<ClientConnectionData> for ExpectServerDone<'_> {
 
             // Check the signature is compatible with the ciphersuite.
             let sig = &st.server_kx.kx_sig;
-            if !SupportedCipherSuite::from(suite).usable_for_signature_algorithm(sig.scheme.sign())
+            if !SupportedCipherSuite::from(suite)
+                .usable_for_signature_algorithm(sig.scheme.algorithm())
             {
                 warn!(
                     "peer signed kx with wrong algorithm (got {:?} expect {:?})",
-                    sig.scheme.sign(),
+                    sig.scheme.algorithm(),
                     suite.sign
                 );
                 return Err(PeerMisbehaved::SignedKxWithWrongAlgorithm.into());
