@@ -10,8 +10,9 @@ use super::client_conn::ClientConnectionData;
 use super::hs::ClientContext;
 use crate::check::inappropriate_handshake_message;
 use crate::client::common::{ClientAuthDetails, ClientHelloDetails, ServerCertDetails};
+use crate::client::ech::{self, EchState, EchStatus};
 use crate::client::{hs, ClientConfig, ClientSessionStore};
-use crate::common_state::{CommonState, Protocol, Side, State};
+use crate::common_state::{CommonState, HandshakeKind, Protocol, Side, State};
 use crate::conn::ConnectionRandoms;
 use crate::crypto::ActiveKeyExchange;
 use crate::enums::{
@@ -23,11 +24,12 @@ use crate::hash_hs::{HandshakeHash, HandshakeHashBuffer};
 use crate::log::{debug, trace, warn};
 use crate::msgs::base::{Payload, PayloadU8};
 use crate::msgs::ccs::ChangeCipherSpecPayload;
+use crate::msgs::codec::{Codec, Reader};
 use crate::msgs::enums::{ExtensionType, KeyUpdateRequest};
 use crate::msgs::handshake::{
-    CertificateEntry, CertificatePayloadTls13, ClientExtension, HandshakeMessagePayload,
+    CertificatePayloadTls13, ClientExtension, EchConfigPayload, HandshakeMessagePayload,
     HandshakePayload, HasServerExtensions, NewSessionTicketPayloadTls13, PresharedKeyIdentity,
-    PresharedKeyOffer, ServerExtension, ServerHelloPayload,
+    PresharedKeyOffer, ServerExtension, ServerHelloPayload, CERTIFICATE_MAX_SIZE_LIMIT,
 };
 use crate::msgs::message::{Message, MessagePayload};
 use crate::msgs::persist;
@@ -40,7 +42,7 @@ use crate::tls13::{
     construct_client_verify_message, construct_server_verify_message, Tls13CipherSuite,
 };
 use crate::verify::{self, DigitallySignedStruct};
-use crate::{crypto, KeyLog};
+use crate::{compress, crypto, KeyLog};
 
 // Extensions we expect in plaintext in the ServerHello.
 static ALLOWED_PLAINTEXT_EXTS: &[ExtensionType] = &[
@@ -64,13 +66,15 @@ pub(super) fn handle_server_hello(
     server_hello: &ServerHelloPayload,
     mut resuming_session: Option<persist::Tls13ClientSessionValue>,
     server_name: ServerName<'static>,
-    randoms: ConnectionRandoms,
+    mut randoms: ConnectionRandoms,
     suite: &'static Tls13CipherSuite,
-    transcript: HandshakeHash,
+    mut transcript: HandshakeHash,
     early_key_schedule: Option<KeyScheduleEarly>,
-    hello: ClientHelloDetails,
+    mut hello: ClientHelloDetails,
     our_key_share: Box<dyn ActiveKeyExchange>,
     mut sent_tls13_fake_ccs: bool,
+    server_hello_msg: &Message,
+    ech_state: Option<EchState>,
 ) -> hs::NextStateOrError<'static> {
     validate_server_hello(cx.common, server_hello)?;
 
@@ -145,7 +149,31 @@ pub(super) fn handle_server_hello(
 
     let shared_secret = our_key_share.complete(&their_key_share.payload.0)?;
 
-    let key_schedule = key_schedule_pre_handshake.into_handshake(shared_secret);
+    let mut key_schedule = key_schedule_pre_handshake.into_handshake(shared_secret);
+
+    // If we have ECH state, check that the server accepted our offer.
+    if let Some(ech_state) = ech_state {
+        cx.data.ech_status = match ech_state.confirm_acceptance(
+            &mut key_schedule,
+            server_hello,
+            suite.common.hash_provider,
+        )? {
+            // The server accepted our ECH offer, so complete the inner transcript with the
+            // server hello message, and switch the relevant state to the copies for the
+            // inner client hello.
+            Some(mut accepted) => {
+                accepted
+                    .transcript
+                    .add_message(server_hello_msg);
+                transcript = accepted.transcript;
+                randoms.client = accepted.random.0;
+                hello.sent_extensions = accepted.sent_extensions;
+                EchStatus::Accepted
+            }
+            // The server rejected our ECH offer.
+            None => EchStatus::Rejected,
+        };
+    }
 
     // Remember what KX group the server liked for next time.
     config
@@ -391,6 +419,22 @@ impl State<ClientConnectionData> for ExpectEncryptedExtensions {
         validate_encrypted_extensions(cx.common, &self.hello, exts)?;
         hs::process_alpn_protocol(cx.common, &self.config, exts.alpn_protocol())?;
 
+        let ech_retry_configs = match (cx.data.ech_status, exts.server_ech_extension()) {
+            // If we didn't offer ECH, or ECH was accepted, but the server sent an ECH encrypted
+            // extension with retry configs, we must error.
+            (EchStatus::NotOffered | EchStatus::Accepted, Some(_)) => {
+                return Err(cx.common.send_fatal_alert(
+                    AlertDescription::UnsupportedExtension,
+                    PeerMisbehaved::UnsolicitedEchExtension,
+                ))
+            }
+            // If we offered ECH, and it was rejected, store the retry configs (if any) from
+            // the server's ECH extension. We will return them in an error produced at the end
+            // of the handshake.
+            (EchStatus::Rejected, ext) => ext.map(|ext| ext.retry_configs.to_vec()),
+            _ => None,
+        };
+
         // QUIC transport parameters
         if cx.common.is_quic() {
             match exts.quic_params_extension() {
@@ -425,6 +469,7 @@ impl State<ClientConnectionData> for ExpectEncryptedExtensions {
                     .server_cert_chain()
                     .clone(),
             );
+            cx.common.handshake_kind = Some(HandshakeKind::Resumed);
 
             // We *don't* reverify the certificate chain here: resumption is a
             // continuation of the previous session in terms of security policy.
@@ -440,19 +485,203 @@ impl State<ClientConnectionData> for ExpectEncryptedExtensions {
                 client_auth: None,
                 cert_verified,
                 sig_verified,
+                ech_retry_configs,
             }))
         } else {
             if exts.early_data_extension_offered() {
                 return Err(PeerMisbehaved::EarlyDataExtensionWithoutResumption.into());
             }
-            Ok(Box::new(ExpectCertificateOrCertReq {
+            cx.common
+                .handshake_kind
+                .get_or_insert(HandshakeKind::Full);
+
+            Ok(if self.hello.offered_cert_compression {
+                Box::new(ExpectCertificateOrCompressedCertificateOrCertReq {
+                    config: self.config,
+                    server_name: self.server_name,
+                    randoms: self.randoms,
+                    suite: self.suite,
+                    transcript: self.transcript,
+                    key_schedule: self.key_schedule,
+                    ech_retry_configs,
+                })
+            } else {
+                Box::new(ExpectCertificateOrCertReq {
+                    config: self.config,
+                    server_name: self.server_name,
+                    randoms: self.randoms,
+                    suite: self.suite,
+                    transcript: self.transcript,
+                    key_schedule: self.key_schedule,
+                    ech_retry_configs,
+                })
+            })
+        }
+    }
+
+    fn into_owned(self: Box<Self>) -> hs::NextState<'static> {
+        self
+    }
+}
+
+struct ExpectCertificateOrCompressedCertificateOrCertReq {
+    config: Arc<ClientConfig>,
+    server_name: ServerName<'static>,
+    randoms: ConnectionRandoms,
+    suite: &'static Tls13CipherSuite,
+    transcript: HandshakeHash,
+    key_schedule: KeyScheduleHandshake,
+    ech_retry_configs: Option<Vec<EchConfigPayload>>,
+}
+
+impl State<ClientConnectionData> for ExpectCertificateOrCompressedCertificateOrCertReq {
+    fn handle<'m>(
+        self: Box<Self>,
+        cx: &mut ClientContext<'_>,
+        m: Message<'m>,
+    ) -> hs::NextStateOrError<'m>
+    where
+        Self: 'm,
+    {
+        match m.payload {
+            MessagePayload::Handshake {
+                parsed:
+                    HandshakeMessagePayload {
+                        payload: HandshakePayload::CertificateTls13(..),
+                        ..
+                    },
+                ..
+            } => Box::new(ExpectCertificate {
                 config: self.config,
                 server_name: self.server_name,
                 randoms: self.randoms,
                 suite: self.suite,
                 transcript: self.transcript,
                 key_schedule: self.key_schedule,
-            }))
+                client_auth: None,
+                message_already_in_transcript: false,
+                ech_retry_configs: self.ech_retry_configs,
+            })
+            .handle(cx, m),
+            MessagePayload::Handshake {
+                parsed:
+                    HandshakeMessagePayload {
+                        payload: HandshakePayload::CompressedCertificate(..),
+                        ..
+                    },
+                ..
+            } => Box::new(ExpectCompressedCertificate {
+                config: self.config,
+                server_name: self.server_name,
+                randoms: self.randoms,
+                suite: self.suite,
+                transcript: self.transcript,
+                key_schedule: self.key_schedule,
+                client_auth: None,
+                ech_retry_configs: self.ech_retry_configs,
+            })
+            .handle(cx, m),
+            MessagePayload::Handshake {
+                parsed:
+                    HandshakeMessagePayload {
+                        payload: HandshakePayload::CertificateRequestTls13(..),
+                        ..
+                    },
+                ..
+            } => Box::new(ExpectCertificateRequest {
+                config: self.config,
+                server_name: self.server_name,
+                randoms: self.randoms,
+                suite: self.suite,
+                transcript: self.transcript,
+                key_schedule: self.key_schedule,
+                offered_cert_compression: true,
+                ech_retry_configs: self.ech_retry_configs,
+            })
+            .handle(cx, m),
+            payload => Err(inappropriate_handshake_message(
+                &payload,
+                &[ContentType::Handshake],
+                &[
+                    HandshakeType::Certificate,
+                    HandshakeType::CertificateRequest,
+                    HandshakeType::CompressedCertificate,
+                ],
+            )),
+        }
+    }
+
+    fn into_owned(self: Box<Self>) -> hs::NextState<'static> {
+        self
+    }
+}
+
+struct ExpectCertificateOrCompressedCertificate {
+    config: Arc<ClientConfig>,
+    server_name: ServerName<'static>,
+    randoms: ConnectionRandoms,
+    suite: &'static Tls13CipherSuite,
+    transcript: HandshakeHash,
+    key_schedule: KeyScheduleHandshake,
+    client_auth: Option<ClientAuthDetails>,
+    ech_retry_configs: Option<Vec<EchConfigPayload>>,
+}
+
+impl State<ClientConnectionData> for ExpectCertificateOrCompressedCertificate {
+    fn handle<'m>(
+        self: Box<Self>,
+        cx: &mut ClientContext<'_>,
+        m: Message<'m>,
+    ) -> hs::NextStateOrError<'m>
+    where
+        Self: 'm,
+    {
+        match m.payload {
+            MessagePayload::Handshake {
+                parsed:
+                    HandshakeMessagePayload {
+                        payload: HandshakePayload::CertificateTls13(..),
+                        ..
+                    },
+                ..
+            } => Box::new(ExpectCertificate {
+                config: self.config,
+                server_name: self.server_name,
+                randoms: self.randoms,
+                suite: self.suite,
+                transcript: self.transcript,
+                key_schedule: self.key_schedule,
+                client_auth: self.client_auth,
+                message_already_in_transcript: false,
+                ech_retry_configs: self.ech_retry_configs,
+            })
+            .handle(cx, m),
+            MessagePayload::Handshake {
+                parsed:
+                    HandshakeMessagePayload {
+                        payload: HandshakePayload::CompressedCertificate(..),
+                        ..
+                    },
+                ..
+            } => Box::new(ExpectCompressedCertificate {
+                config: self.config,
+                server_name: self.server_name,
+                randoms: self.randoms,
+                suite: self.suite,
+                transcript: self.transcript,
+                key_schedule: self.key_schedule,
+                client_auth: self.client_auth,
+                ech_retry_configs: self.ech_retry_configs,
+            })
+            .handle(cx, m),
+            payload => Err(inappropriate_handshake_message(
+                &payload,
+                &[ContentType::Handshake],
+                &[
+                    HandshakeType::Certificate,
+                    HandshakeType::CompressedCertificate,
+                ],
+            )),
         }
     }
 
@@ -468,6 +697,7 @@ struct ExpectCertificateOrCertReq {
     suite: &'static Tls13CipherSuite,
     transcript: HandshakeHash,
     key_schedule: KeyScheduleHandshake,
+    ech_retry_configs: Option<Vec<EchConfigPayload>>,
 }
 
 impl State<ClientConnectionData> for ExpectCertificateOrCertReq {
@@ -495,6 +725,8 @@ impl State<ClientConnectionData> for ExpectCertificateOrCertReq {
                 transcript: self.transcript,
                 key_schedule: self.key_schedule,
                 client_auth: None,
+                message_already_in_transcript: false,
+                ech_retry_configs: self.ech_retry_configs,
             })
             .handle(cx, m),
             MessagePayload::Handshake {
@@ -511,6 +743,8 @@ impl State<ClientConnectionData> for ExpectCertificateOrCertReq {
                 suite: self.suite,
                 transcript: self.transcript,
                 key_schedule: self.key_schedule,
+                offered_cert_compression: false,
+                ech_retry_configs: self.ech_retry_configs,
             })
             .handle(cx, m),
             payload => Err(inappropriate_handshake_message(
@@ -539,6 +773,8 @@ struct ExpectCertificateRequest {
     suite: &'static Tls13CipherSuite,
     transcript: HandshakeHash,
     key_schedule: KeyScheduleHandshake,
+    offered_cert_compression: bool,
+    ech_retry_configs: Option<Vec<EchConfigPayload>>,
 }
 
 impl State<ClientConnectionData> for ExpectCertificateRequest {
@@ -586,6 +822,16 @@ impl State<ClientConnectionData> for ExpectCertificateRequest {
             ));
         }
 
+        let compat_compressor = certreq
+            .certificate_compression_extension()
+            .and_then(|offered| {
+                self.config
+                    .cert_compressors
+                    .iter()
+                    .find(|compressor| offered.contains(&compressor.algorithm()))
+            })
+            .cloned();
+
         let client_auth = ClientAuthDetails::resolve(
             self.config
                 .client_auth_cert_resolver
@@ -593,17 +839,139 @@ impl State<ClientConnectionData> for ExpectCertificateRequest {
             certreq.authorities_extension(),
             &compat_sigschemes,
             Some(certreq.context.0.clone()),
+            compat_compressor,
         );
 
-        Ok(Box::new(ExpectCertificate {
+        Ok(if self.offered_cert_compression {
+            Box::new(ExpectCertificateOrCompressedCertificate {
+                config: self.config,
+                server_name: self.server_name,
+                randoms: self.randoms,
+                suite: self.suite,
+                transcript: self.transcript,
+                key_schedule: self.key_schedule,
+                client_auth: Some(client_auth),
+                ech_retry_configs: self.ech_retry_configs,
+            })
+        } else {
+            Box::new(ExpectCertificate {
+                config: self.config,
+                server_name: self.server_name,
+                randoms: self.randoms,
+                suite: self.suite,
+                transcript: self.transcript,
+                key_schedule: self.key_schedule,
+                client_auth: Some(client_auth),
+                message_already_in_transcript: false,
+                ech_retry_configs: self.ech_retry_configs,
+            })
+        })
+    }
+
+    fn into_owned(self: Box<Self>) -> hs::NextState<'static> {
+        self
+    }
+}
+
+struct ExpectCompressedCertificate {
+    config: Arc<ClientConfig>,
+    server_name: ServerName<'static>,
+    randoms: ConnectionRandoms,
+    suite: &'static Tls13CipherSuite,
+    transcript: HandshakeHash,
+    key_schedule: KeyScheduleHandshake,
+    client_auth: Option<ClientAuthDetails>,
+    ech_retry_configs: Option<Vec<EchConfigPayload>>,
+}
+
+impl State<ClientConnectionData> for ExpectCompressedCertificate {
+    fn handle<'m>(
+        mut self: Box<Self>,
+        cx: &mut ClientContext<'_>,
+        m: Message<'m>,
+    ) -> hs::NextStateOrError<'m>
+    where
+        Self: 'm,
+    {
+        self.transcript.add_message(&m);
+        let compressed_cert = require_handshake_msg_move!(
+            m,
+            HandshakeType::CompressedCertificate,
+            HandshakePayload::CompressedCertificate
+        )?;
+
+        let decompressor = match self
+            .config
+            .cert_decompressors
+            .iter()
+            .find(|item| item.algorithm() == compressed_cert.alg)
+        {
+            Some(dec) => dec,
+            None => {
+                return Err(cx.common.send_fatal_alert(
+                    AlertDescription::BadCertificate,
+                    PeerMisbehaved::SelectedUnofferedCertCompression,
+                ));
+            }
+        };
+
+        if compressed_cert.uncompressed_len as usize > CERTIFICATE_MAX_SIZE_LIMIT {
+            return Err(cx.common.send_fatal_alert(
+                AlertDescription::BadCertificate,
+                InvalidMessage::MessageTooLarge,
+            ));
+        }
+
+        let mut decompress_buffer = vec![0u8; compressed_cert.uncompressed_len as usize];
+        if let Err(compress::DecompressionFailed) =
+            decompressor.decompress(compressed_cert.compressed.0.bytes(), &mut decompress_buffer)
+        {
+            return Err(cx.common.send_fatal_alert(
+                AlertDescription::BadCertificate,
+                PeerMisbehaved::InvalidCertCompression,
+            ));
+        }
+
+        let cert_payload =
+            match CertificatePayloadTls13::read(&mut Reader::init(&decompress_buffer)) {
+                Ok(cm) => cm,
+                Err(err) => {
+                    return Err(cx
+                        .common
+                        .send_fatal_alert(AlertDescription::BadCertificate, err));
+                }
+            };
+        trace!(
+            "Server certificate decompressed using {:?} ({} bytes -> {})",
+            compressed_cert.alg,
+            compressed_cert
+                .compressed
+                .0
+                .bytes()
+                .len(),
+            compressed_cert.uncompressed_len,
+        );
+
+        let m = Message {
+            version: ProtocolVersion::TLSv1_3,
+            payload: MessagePayload::handshake(HandshakeMessagePayload {
+                typ: HandshakeType::Certificate,
+                payload: HandshakePayload::CertificateTls13(cert_payload.into_owned()),
+            }),
+        };
+
+        Box::new(ExpectCertificate {
             config: self.config,
             server_name: self.server_name,
             randoms: self.randoms,
             suite: self.suite,
             transcript: self.transcript,
             key_schedule: self.key_schedule,
-            client_auth: Some(client_auth),
-        }))
+            client_auth: self.client_auth,
+            message_already_in_transcript: true,
+            ech_retry_configs: self.ech_retry_configs,
+        })
+        .handle(cx, m)
     }
 
     fn into_owned(self: Box<Self>) -> hs::NextState<'static> {
@@ -619,6 +987,8 @@ struct ExpectCertificate {
     transcript: HandshakeHash,
     key_schedule: KeyScheduleHandshake,
     client_auth: Option<ClientAuthDetails>,
+    message_already_in_transcript: bool,
+    ech_retry_configs: Option<Vec<EchConfigPayload>>,
 }
 
 impl State<ClientConnectionData> for ExpectCertificate {
@@ -630,7 +1000,9 @@ impl State<ClientConnectionData> for ExpectCertificate {
     where
         Self: 'm,
     {
-        self.transcript.add_message(&m);
+        if !self.message_already_in_transcript {
+            self.transcript.add_message(&m);
+        }
         let cert_chain = require_handshake_msg_move!(
             m,
             HandshakeType::Certificate,
@@ -654,7 +1026,12 @@ impl State<ClientConnectionData> for ExpectCertificate {
             ));
         }
         let end_entity_ocsp = cert_chain.end_entity_ocsp();
-        let server_cert = ServerCertDetails::new(cert_chain.convert(), end_entity_ocsp);
+        let server_cert = ServerCertDetails::new(
+            cert_chain
+                .into_certificate_chain()
+                .into_owned(),
+            end_entity_ocsp,
+        );
 
         Ok(Box::new(ExpectCertificateVerify {
             config: self.config,
@@ -665,6 +1042,7 @@ impl State<ClientConnectionData> for ExpectCertificate {
             key_schedule: self.key_schedule,
             server_cert,
             client_auth: self.client_auth,
+            ech_retry_configs: self.ech_retry_configs,
         }))
     }
 
@@ -683,6 +1061,7 @@ struct ExpectCertificateVerify<'a> {
     key_schedule: KeyScheduleHandshake,
     server_cert: ServerCertDetails<'a>,
     client_auth: Option<ClientAuthDetails>,
+    ech_retry_configs: Option<Vec<EchConfigPayload>>,
 }
 
 impl State<ClientConnectionData> for ExpectCertificateVerify<'_> {
@@ -754,6 +1133,7 @@ impl State<ClientConnectionData> for ExpectCertificateVerify<'_> {
             client_auth: self.client_auth,
             cert_verified,
             sig_verified,
+            ech_retry_configs: self.ech_retry_configs,
         }))
     }
 
@@ -767,30 +1147,52 @@ impl State<ClientConnectionData> for ExpectCertificateVerify<'_> {
             key_schedule: self.key_schedule,
             server_cert: self.server_cert.into_owned(),
             client_auth: self.client_auth,
+            ech_retry_configs: self.ech_retry_configs,
         })
     }
 }
 
+fn emit_compressed_certificate_tls13(
+    transcript: &mut HandshakeHash,
+    certkey: &CertifiedKey,
+    auth_context: Option<Vec<u8>>,
+    compressor: &dyn compress::CertCompressor,
+    config: &ClientConfig,
+    common: &mut CommonState,
+) {
+    let mut cert_payload = CertificatePayloadTls13::new(certkey.cert.iter(), None);
+    cert_payload.context = PayloadU8::new(auth_context.clone().unwrap_or_default());
+
+    let compressed = match config
+        .cert_compression_cache
+        .compression_for(compressor, &cert_payload)
+    {
+        Ok(compressed) => compressed,
+        Err(_) => return emit_certificate_tls13(transcript, Some(certkey), auth_context, common),
+    };
+
+    let m = Message {
+        version: ProtocolVersion::TLSv1_3,
+        payload: MessagePayload::handshake(HandshakeMessagePayload {
+            typ: HandshakeType::CompressedCertificate,
+            payload: HandshakePayload::CompressedCertificate(compressed.compressed_cert_payload()),
+        }),
+    };
+    transcript.add_message(&m);
+    common.send_msg(m, true);
+}
+
 fn emit_certificate_tls13(
     transcript: &mut HandshakeHash,
     certkey: Option<&CertifiedKey>,
     auth_context: Option<Vec<u8>>,
     common: &mut CommonState,
 ) {
-    let context = auth_context.unwrap_or_default();
-
-    let mut cert_payload = CertificatePayloadTls13 {
-        context: PayloadU8::new(context),
-        entries: Vec::new(),
-    };
-
-    if let Some(certkey) = certkey {
-        for cert in &certkey.cert {
-            cert_payload
-                .entries
-                .push(CertificateEntry::new(cert.clone()));
-        }
-    }
+    let certs = certkey
+        .map(|ck| ck.cert.as_ref())
+        .unwrap_or(&[][..]);
+    let mut cert_payload = CertificatePayloadTls13::new(certs.iter(), None);
+    cert_payload.context = PayloadU8::new(auth_context.unwrap_or_default());
 
     let m = Message {
         version: ProtocolVersion::TLSv1_3,
@@ -873,6 +1275,7 @@ struct ExpectFinished {
     client_auth: Option<ClientAuthDetails>,
     cert_verified: verify::ServerCertVerified,
     sig_verified: verify::HandshakeSignatureValid,
+    ech_retry_configs: Option<Vec<EchConfigPayload>>,
 }
 
 impl State<ClientConnectionData> for ExpectFinished {
@@ -925,17 +1328,37 @@ impl State<ClientConnectionData> for ExpectFinished {
                 } => {
                     emit_certificate_tls13(&mut st.transcript, None, auth_context, cx.common);
                 }
+                ClientAuthDetails::Verify {
+                    auth_context_tls13: auth_context,
+                    ..
+                } if cx.data.ech_status == EchStatus::Rejected => {
+                    // If ECH was offered, and rejected, we MUST respond with
+                    // an empty certificate message.
+                    emit_certificate_tls13(&mut st.transcript, None, auth_context, cx.common);
+                }
                 ClientAuthDetails::Verify {
                     certkey,
                     signer,
                     auth_context_tls13: auth_context,
+                    compressor,
                 } => {
-                    emit_certificate_tls13(
-                        &mut st.transcript,
-                        Some(&certkey),
-                        auth_context,
-                        cx.common,
-                    );
+                    if let Some(compressor) = compressor {
+                        emit_compressed_certificate_tls13(
+                            &mut st.transcript,
+                            &certkey,
+                            auth_context,
+                            compressor,
+                            &st.config,
+                            cx.common,
+                        );
+                    } else {
+                        emit_certificate_tls13(
+                            &mut st.transcript,
+                            Some(&certkey),
+                            auth_context,
+                            cx.common,
+                        );
+                    }
                     emit_certverify_tls13(&mut st.transcript, signer.as_ref(), cx.common)?;
                 }
             }
@@ -965,6 +1388,13 @@ impl State<ClientConnectionData> for ExpectFinished {
         cx.common
             .start_traffic(&mut cx.sendable_plaintext);
 
+        // Now that we've reached the end of the normal handshake we must enforce ECH acceptance by
+        // sending an alert and returning an error (potentially with retry configs) if the server
+        // did not accept our ECH offer.
+        if cx.data.ech_status == EchStatus::Rejected {
+            return Err(ech::fatal_alert_required(st.ech_retry_configs, cx.common));
+        }
+
         let st = ExpectTraffic {
             config: Arc::clone(&st.config),
             session_storage: Arc::clone(&st.config.resumption.store),
