@@ -26,6 +26,7 @@ use crate::{quic, record_layer};
 /// Connection state common to both client and server connections.
 pub struct CommonState {
     pub(crate) negotiated_version: Option<ProtocolVersion>,
+    pub(crate) handshake_kind: Option<HandshakeKind>,
     pub(crate) side: Side,
     pub(crate) record_layer: record_layer::RecordLayer,
     pub(crate) suite: Option<SupportedCipherSuite>,
@@ -50,12 +51,14 @@ pub struct CommonState {
     pub(crate) protocol: Protocol,
     pub(crate) quic: quic::Quic,
     pub(crate) enable_secret_extraction: bool,
+    temper_counters: TemperCounters,
 }
 
 impl CommonState {
     pub(crate) fn new(side: Side) -> Self {
         Self {
             negotiated_version: None,
+            handshake_kind: None,
             side,
             record_layer: record_layer::RecordLayer::new(),
             suite: None,
@@ -77,6 +80,7 @@ impl CommonState {
             protocol: Protocol::Tcp,
             quic: quic::Quic::default(),
             enable_secret_extraction: false,
+            temper_counters: TemperCounters::default(),
         }
     }
 
@@ -140,6 +144,16 @@ impl CommonState {
         self.negotiated_version
     }
 
+    /// Which kind of handshake was performed.
+    ///
+    /// This tells you whether the handshake was a resumption or not.
+    ///
+    /// This will return `None` before it is known which sort of
+    /// handshake occurred.
+    pub fn handshake_kind(&self) -> Option<HandshakeKind> {
+        self.handshake_kind
+    }
+
     pub(crate) fn is_tls13(&self) -> bool {
         matches!(self.negotiated_version, Some(ProtocolVersion::TLSv1_3))
     }
@@ -159,6 +173,8 @@ impl CommonState {
                 Side::Server => HandshakeType::ClientHello,
             };
             if msg.is_handshake_type(reject_ty) {
+                self.temper_counters
+                    .received_renegotiation_request()?;
                 self.send_warning_alert(AlertDescription::NoRenegotiation);
                 return Ok(state);
             }
@@ -419,8 +435,8 @@ impl CommonState {
         }
 
         // If we get a CloseNotify, make a note to declare EOF to our
-        // caller.
-        if alert.description == AlertDescription::CloseNotify {
+        // caller.  But do not treat unauthenticated alerts like this.
+        if self.may_receive_application_data && alert.description == AlertDescription::CloseNotify {
             self.has_received_close_notify = true;
             return Ok(());
         }
@@ -429,6 +445,8 @@ impl CommonState {
         // (except, for no good reason, user_cancelled).
         let err = Error::AlertReceived(alert.description);
         if alert.level == AlertLevel::Warning {
+            self.temper_counters
+                .received_warning_alert()?;
             if self.is_tls13() && alert.description != AlertDescription::UserCanceled {
                 return Err(self.send_fatal_alert(AlertDescription::DecodeError, err));
             } else {
@@ -463,14 +481,20 @@ impl CommonState {
         err.into()
     }
 
-    /// Queues a close_notify warning alert to be sent in the next
+    /// Queues a `close_notify` warning alert to be sent in the next
     /// [`Connection::write_tls`] call.  This informs the peer that the
     /// connection is being closed.
     ///
+    /// Does nothing if any `close_notify` or fatal alert was already sent.
+    ///
     /// [`Connection::write_tls`]: crate::Connection::write_tls
     pub fn send_close_notify(&mut self) {
+        if self.sent_fatal_alert {
+            return;
+        }
         debug!("Sending warning alert {:?}", AlertDescription::CloseNotify);
         self.send_warning_alert_no_log(AlertDescription::CloseNotify);
+        self.sent_fatal_alert = true;
     }
 
     pub(crate) fn eager_send_close_notify(
@@ -603,6 +627,9 @@ impl CommonState {
         &mut self,
         key_update_request: &KeyUpdateRequest,
     ) -> Result<bool, Error> {
+        self.temper_counters
+            .received_key_update_request()?;
+
         match key_update_request {
             KeyUpdateRequest::UpdateNotRequested => Ok(false),
             KeyUpdateRequest::UpdateRequested => Ok(self.queued_key_update_message.is_none()),
@@ -682,6 +709,30 @@ impl CommonState {
     }
 }
 
+/// Describes which sort of handshake happened.
+#[derive(Debug, PartialEq, Clone, Copy)]
+pub enum HandshakeKind {
+    /// A full handshake.
+    ///
+    /// This is the typical TLS connection initiation process when resumption is
+    /// not yet unavailable, and the initial `ClientHello` was accepted by the server.
+    Full,
+
+    /// A full TLS1.3 handshake, with an extra round-trip for a `HelloRetryRequest`.
+    ///
+    /// The server can respond with a `HelloRetryRequest` if the initial `ClientHello`
+    /// is unacceptable for several reasons, the most likely if no supported key
+    /// shares were offered by the client.
+    FullWithHelloRetryRequest,
+
+    /// A resumed handshake.
+    ///
+    /// Resumed handshakes involve fewer round trips and less cryptography than
+    /// full ones, but can only happen when the peers have previously done a full
+    /// handshake together, and then remember data about it.
+    Resumed,
+}
+
 /// Values of this structure are returned from [`Connection::process_new_packets`]
 /// and tell the caller the current I/O state of the TLS connection.
 ///
@@ -786,5 +837,63 @@ enum Limit {
     No,
 }
 
+/// Tracking technically-allowed protocol actions
+/// that we limit to avoid denial-of-service vectors.
+struct TemperCounters {
+    allowed_warning_alerts: u8,
+    allowed_renegotiation_requests: u8,
+    allowed_key_update_requests: u8,
+}
+
+impl TemperCounters {
+    fn received_warning_alert(&mut self) -> Result<(), Error> {
+        match self.allowed_warning_alerts {
+            0 => Err(PeerMisbehaved::TooManyWarningAlertsReceived.into()),
+            _ => {
+                self.allowed_warning_alerts -= 1;
+                Ok(())
+            }
+        }
+    }
+
+    fn received_renegotiation_request(&mut self) -> Result<(), Error> {
+        match self.allowed_renegotiation_requests {
+            0 => Err(PeerMisbehaved::TooManyRenegotiationRequests.into()),
+            _ => {
+                self.allowed_renegotiation_requests -= 1;
+                Ok(())
+            }
+        }
+    }
+
+    fn received_key_update_request(&mut self) -> Result<(), Error> {
+        match self.allowed_key_update_requests {
+            0 => Err(PeerMisbehaved::TooManyKeyUpdateRequests.into()),
+            _ => {
+                self.allowed_key_update_requests -= 1;
+                Ok(())
+            }
+        }
+    }
+}
+
+impl Default for TemperCounters {
+    fn default() -> Self {
+        Self {
+            // cf. BoringSSL `kMaxWarningAlerts`
+            // <https://github.com/google/boringssl/blob/dec5989b793c56ad4dd32173bd2d8595ca78b398/ssl/tls_record.cc#L137-L139>
+            allowed_warning_alerts: 4,
+
+            // we rebuff renegotiation requests with a `NoRenegotiation` warning alerts.
+            // a second request after this is fatal.
+            allowed_renegotiation_requests: 1,
+
+            // cf. BoringSSL `kMaxKeyUpdates`
+            // <https://github.com/google/boringssl/blob/dec5989b793c56ad4dd32173bd2d8595ca78b398/ssl/tls13_both.cc#L35-L38>
+            allowed_key_update_requests: 32,
+        }
+    }
+}
+
 const DEFAULT_RECEIVED_PLAINTEXT_LIMIT: usize = 16 * 1024;
 pub(crate) const DEFAULT_BUFFER_LIMIT: usize = 64 * 1024;
