@@ -509,7 +509,7 @@ impl ExpectClientHello {
             .find_map(|maybe_skxg| match maybe_skxg {
                 Some(skxg) => suite
                     .usable_for_kx_algorithm(skxg.name().key_exchange_algorithm())
-                    .then(|| *skxg),
+                    .then_some(*skxg),
                 None => None,
             });
 
@@ -640,7 +640,6 @@ pub(super) fn process_client_hello<'a>(
     Ok((client_hello, sig_schemes.to_owned()))
 }
 
-#[allow(clippy::large_enum_variant)]
 pub(crate) enum HandshakeHashOrBuffer {
     Buffer(HandshakeHashBuffer),
     Hash(HandshakeHash),
