@@ -32,7 +32,7 @@ use crate::time_provider::TimeProvider;
 use crate::vecbuf::ChunkVecBuffer;
 #[cfg(feature = "std")]
 use crate::WantsVerifier;
-use crate::{sign, verify, versions, KeyLog, WantsVersions};
+use crate::{compress, sign, verify, versions, KeyLog, WantsVersions};
 
 /// A trait for the ability to store server session data.
 ///
@@ -212,13 +212,19 @@ impl<'a> ClientHello<'a> {
 /// * [`ServerConfig::max_fragment_size`]: the default is `None` (meaning 16kB).
 /// * [`ServerConfig::session_storage`]: if the `std` feature is enabled, the default stores 256
 ///   sessions in memory. If the `std` feature is not enabled, the default is to not store any
-///   sessions.
+///   sessions. In a no-std context, by enabling the `hashbrown` feature you may provide your
+///   own `session_storage` using [`ServerSessionMemoryCache`] and a `crate::lock::MakeMutex`
+///   implementation.
 /// * [`ServerConfig::alpn_protocols`]: the default is empty -- no ALPN protocol is negotiated.
 /// * [`ServerConfig::key_log`]: key material is not logged.
 /// * [`ServerConfig::send_tls13_tickets`]: 4 tickets are sent.
+/// * [`ServerConfig::cert_compressors`]: depends on the crate features, see [`compress::default_cert_compressors()`].
+/// * [`ServerConfig::cert_compression_cache`]: caches the most recently used 4 compressions
+/// * [`ServerConfig::cert_decompressors`]: depends on the crate features, see [`compress::default_cert_decompressors()`].
 ///
 /// [`RootCertStore`]: crate::RootCertStore
-#[derive(Debug)]
+/// [`ServerSessionMemoryCache`]: crate::server::handy::ServerSessionMemoryCache
+#[derive(Clone, Debug)]
 pub struct ServerConfig {
     /// Source of randomness and other crypto.
     pub(super) provider: Arc<CryptoProvider>,
@@ -334,31 +340,37 @@ pub struct ServerConfig {
 
     /// Provides the current system time
     pub time_provider: Arc<dyn TimeProvider>,
-}
 
-// Avoid a `Clone` bound on `C`.
-impl Clone for ServerConfig {
-    fn clone(&self) -> Self {
-        Self {
-            provider: Arc::<CryptoProvider>::clone(&self.provider),
-            ignore_client_order: self.ignore_client_order,
-            max_fragment_size: self.max_fragment_size,
-            session_storage: Arc::clone(&self.session_storage),
-            ticketer: Arc::clone(&self.ticketer),
-            cert_resolver: Arc::clone(&self.cert_resolver),
-            alpn_protocols: self.alpn_protocols.clone(),
-            versions: self.versions,
-            verifier: Arc::clone(&self.verifier),
-            key_log: Arc::clone(&self.key_log),
-            enable_secret_extraction: self.enable_secret_extraction,
-            max_early_data_size: self.max_early_data_size,
-            send_half_rtt_data: self.send_half_rtt_data,
-            send_tls13_tickets: self.send_tls13_tickets,
-            #[cfg(feature = "tls12")]
-            require_ems: self.require_ems,
-            time_provider: Arc::clone(&self.time_provider),
-        }
-    }
+    /// How to compress the server's certificate chain.
+    ///
+    /// If a client supports this extension, and advertises support
+    /// for one of the compression algorithms included here, the
+    /// server certificate will be compressed according to [RFC8779].
+    ///
+    /// This only applies to TLS1.3 connections.  It is ignored for
+    /// TLS1.2 connections.
+    ///
+    /// [RFC8779]: https://datatracker.ietf.org/doc/rfc8879/
+    pub cert_compressors: Vec<&'static dyn compress::CertCompressor>,
+
+    /// Caching for compressed certificates.
+    ///
+    /// This is optional: [`compress::CompressionCache::Disabled`] gives
+    /// a cache that does no caching.
+    pub cert_compression_cache: Arc<compress::CompressionCache>,
+
+    /// How to decompress the clients's certificate chain.
+    ///
+    /// If this is non-empty, the [RFC8779] certificate compression
+    /// extension is offered when requesting client authentication,
+    /// and any compressed certificates are transparently decompressed
+    /// during the handshake.
+    ///
+    /// This only applies to TLS1.3 connections.  It is ignored for
+    /// TLS1.2 connections.
+    ///
+    /// [RFC8779]: https://datatracker.ietf.org/doc/rfc8879/
+    pub cert_decompressors: Vec<&'static dyn compress::CertDecompressor>,
 }
 
 impl ServerConfig {
