@@ -1,4 +1,3 @@
-use alloc::borrow::ToOwned;
 use alloc::boxed::Box;
 use alloc::sync::Arc;
 use alloc::vec;
@@ -11,18 +10,18 @@ use subtle::ConstantTimeEq;
 use super::hs::{self, HandshakeHashOrBuffer, ServerContext};
 use super::server_conn::ServerConnectionData;
 use crate::check::{inappropriate_handshake_message, inappropriate_message};
-use crate::common_state::{CommonState, Protocol, Side, State};
+use crate::common_state::{CommonState, HandshakeKind, Protocol, Side, State};
 use crate::conn::ConnectionRandoms;
 use crate::enums::{AlertDescription, ContentType, HandshakeType, ProtocolVersion};
-use crate::error::{Error, PeerIncompatible, PeerMisbehaved};
+use crate::error::{Error, InvalidMessage, PeerIncompatible, PeerMisbehaved};
 use crate::hash_hs::HandshakeHash;
 #[cfg(feature = "logging")]
 use crate::log::{debug, trace, warn};
-use crate::msgs::codec::Codec;
+use crate::msgs::codec::{Codec, Reader};
 use crate::msgs::enums::KeyUpdateRequest;
 use crate::msgs::handshake::{
-    CertificateChain, HandshakeMessagePayload, HandshakePayload, NewSessionTicketExtension,
-    NewSessionTicketPayloadTls13,
+    CertificateChain, CertificatePayloadTls13, HandshakeMessagePayload, HandshakePayload,
+    NewSessionTicketExtension, NewSessionTicketPayloadTls13, CERTIFICATE_MAX_SIZE_LIMIT,
 };
 use crate::msgs::message::{Message, MessagePayload};
 use crate::msgs::persist;
@@ -32,19 +31,20 @@ use crate::tls13::key_schedule::{KeyScheduleTraffic, KeyScheduleTrafficWithClien
 use crate::tls13::{
     construct_client_verify_message, construct_server_verify_message, Tls13CipherSuite,
 };
-use crate::{rand, verify};
+use crate::{compress, rand, verify};
 
 mod client_hello {
     use super::*;
+    use crate::compress::CertCompressor;
     use crate::crypto::SupportedKxGroup;
     use crate::enums::SignatureScheme;
     use crate::msgs::base::{Payload, PayloadU8};
     use crate::msgs::ccs::ChangeCipherSpecPayload;
     use crate::msgs::enums::{Compression, NamedGroup, PSKKeyExchangeMode};
     use crate::msgs::handshake::{
-        CertReqExtension, CertificateEntry, CertificateExtension, CertificatePayloadTls13,
-        CertificateRequestPayloadTls13, CertificateStatus, ClientHelloPayload, HelloRetryExtension,
-        HelloRetryRequest, KeyShareEntry, Random, ServerExtension, ServerHelloPayload, SessionId,
+        CertReqExtension, CertificatePayloadTls13, CertificateRequestPayloadTls13,
+        ClientHelloPayload, HelloRetryExtension, HelloRetryRequest, KeyShareEntry, Random,
+        ServerExtension, ServerHelloPayload, SessionId,
     };
     use crate::server::common::ActiveCertifiedKey;
     use crate::sign;
@@ -160,6 +160,24 @@ mod client_hello {
                 ));
             }
 
+            if client_hello.has_certificate_compression_extension_with_duplicates() {
+                return Err(cx.common.send_fatal_alert(
+                    AlertDescription::IllegalParameter,
+                    PeerMisbehaved::OfferedDuplicateCertificateCompressions,
+                ));
+            }
+
+            let cert_compressor = client_hello
+                .certificate_compression_extension()
+                .and_then(|offered|
+                    // prefer server order when choosing a compression: the client's
+                    // extension here does not denote any preference.
+                    self.config
+                        .cert_compressors
+                        .iter()
+                        .find(|compressor| offered.contains(&compressor.algorithm()))
+                        .cloned());
+
             let early_data_requested = client_hello.early_data_extension_offered();
 
             // EarlyData extension is illegal in second ClientHello
@@ -174,7 +192,7 @@ mod client_hello {
 
             // See if there is a KeyShare for the selected kx group.
             let chosen_share_and_kxg = shares_ext.iter().find_map(|share| {
-                (share.group == selected_kxg.name()).then(|| (share, selected_kxg))
+                (share.group == selected_kxg.name()).then_some((share, selected_kxg))
             });
 
             let chosen_share_and_kxg = match chosen_share_and_kxg {
@@ -329,6 +347,14 @@ mod client_hello {
                 emit_fake_ccs(cx.common);
             }
 
+            if full_handshake {
+                cx.common
+                    .handshake_kind
+                    .get_or_insert(HandshakeKind::Full);
+            } else {
+                cx.common.handshake_kind = Some(HandshakeKind::Resumed);
+            }
+
             let mut ocsp_response = server_key.get_ocsp();
             let doing_early_data = emit_encrypted_extensions(
                 &mut self.transcript,
@@ -344,12 +370,24 @@ mod client_hello {
             let doing_client_auth = if full_handshake {
                 let client_auth =
                     emit_certificate_req_tls13(&mut self.transcript, cx, &self.config)?;
-                emit_certificate_tls13(
-                    &mut self.transcript,
-                    cx.common,
-                    server_key.get_cert(),
-                    ocsp_response,
-                );
+
+                if let Some(compressor) = cert_compressor {
+                    emit_compressed_certificate_tls13(
+                        &mut self.transcript,
+                        cx.common,
+                        &self.config,
+                        server_key.get_cert(),
+                        ocsp_response,
+                        compressor,
+                    );
+                } else {
+                    emit_certificate_tls13(
+                        &mut self.transcript,
+                        cx.common,
+                        server_key.get_cert(),
+                        ocsp_response,
+                    );
+                }
                 emit_certificate_verify_tls13(
                     &mut self.transcript,
                     cx.common,
@@ -401,13 +439,28 @@ mod client_hello {
             }
 
             if doing_client_auth {
-                Ok(Box::new(ExpectCertificate {
-                    config: self.config,
-                    transcript: self.transcript,
-                    suite: self.suite,
-                    key_schedule: key_schedule_traffic,
-                    send_tickets: self.send_tickets,
-                }))
+                if self
+                    .config
+                    .cert_decompressors
+                    .is_empty()
+                {
+                    Ok(Box::new(ExpectCertificate {
+                        config: self.config,
+                        transcript: self.transcript,
+                        suite: self.suite,
+                        key_schedule: key_schedule_traffic,
+                        send_tickets: self.send_tickets,
+                        message_already_in_transcript: false,
+                    }))
+                } else {
+                    Ok(Box::new(ExpectCertificateOrCompressedCertificate {
+                        config: self.config,
+                        transcript: self.transcript,
+                        suite: self.suite,
+                        key_schedule: key_schedule_traffic,
+                        send_tickets: self.send_tickets,
+                    }))
+                }
             } else if doing_early_data == EarlyDataDecision::Accepted && !cx.common.is_quic() {
                 // Not used for QUIC: RFC 9001 §8.3: Clients MUST NOT send the EndOfEarlyData
                 // message. A server MUST treat receipt of a CRYPTO frame in a 0-RTT packet as a
@@ -555,6 +608,7 @@ mod client_hello {
         transcript.rollup_for_hrr();
         transcript.add_message(&m);
         common.send_msg(m, false);
+        common.handshake_kind = Some(HandshakeKind::FullWithHelloRetryRequest);
     }
 
     fn decide_if_early_data_allowed(
@@ -668,6 +722,17 @@ mod client_hello {
         cr.extensions
             .push(CertReqExtension::SignatureAlgorithms(schemes.to_vec()));
 
+        if !config.cert_decompressors.is_empty() {
+            cr.extensions
+                .push(CertReqExtension::CertificateCompressionAlgorithms(
+                    config
+                        .cert_decompressors
+                        .iter()
+                        .map(|decomp| decomp.algorithm())
+                        .collect(),
+                ));
+        }
+
         let authorities = config.verifier.root_hint_subjects();
         if !authorities.is_empty() {
             cr.extensions
@@ -694,28 +759,7 @@ mod client_hello {
         cert_chain: &[CertificateDer<'static>],
         ocsp_response: Option<&[u8]>,
     ) {
-        let mut cert_entries = vec![];
-        for cert in cert_chain {
-            let entry = CertificateEntry {
-                cert: cert.to_owned(),
-                exts: Vec::new(),
-            };
-
-            cert_entries.push(entry);
-        }
-
-        if let Some(end_entity_cert) = cert_entries.first_mut() {
-            // Apply OCSP response to first certificate (we don't support OCSP
-            // except for leaf certs).
-            if let Some(ocsp) = ocsp_response {
-                let cst = CertificateStatus::new(ocsp.to_owned());
-                end_entity_cert
-                    .exts
-                    .push(CertificateExtension::CertificateStatus(cst));
-            }
-        }
-
-        let cert_body = CertificatePayloadTls13::new(cert_entries);
+        let cert_body = CertificatePayloadTls13::new(cert_chain.iter(), ocsp_response);
         let c = Message {
             version: ProtocolVersion::TLSv1_3,
             payload: MessagePayload::handshake(HandshakeMessagePayload {
@@ -729,6 +773,37 @@ mod client_hello {
         common.send_msg(c, true);
     }
 
+    fn emit_compressed_certificate_tls13(
+        transcript: &mut HandshakeHash,
+        common: &mut CommonState,
+        config: &ServerConfig,
+        cert_chain: &[CertificateDer<'static>],
+        ocsp_response: Option<&[u8]>,
+        cert_compressor: &'static dyn CertCompressor,
+    ) {
+        let payload = CertificatePayloadTls13::new(cert_chain.iter(), ocsp_response);
+
+        let entry = match config
+            .cert_compression_cache
+            .compression_for(cert_compressor, &payload)
+        {
+            Ok(entry) => entry,
+            Err(_) => return emit_certificate_tls13(transcript, common, cert_chain, ocsp_response),
+        };
+
+        let c = Message {
+            version: ProtocolVersion::TLSv1_3,
+            payload: MessagePayload::handshake(HandshakeMessagePayload {
+                typ: HandshakeType::CompressedCertificate,
+                payload: HandshakePayload::CompressedCertificate(entry.compressed_cert_payload()),
+            }),
+        };
+
+        trace!("sending compressed certificate {:?}", c);
+        transcript.add_message(&c);
+        common.send_msg(c, true);
+    }
+
     fn emit_certificate_verify_tls13(
         transcript: &mut HandshakeHash,
         common: &mut CommonState,
@@ -833,12 +908,180 @@ impl State<ServerConnectionData> for ExpectAndSkipRejectedEarlyData {
     }
 }
 
+struct ExpectCertificateOrCompressedCertificate {
+    config: Arc<ServerConfig>,
+    transcript: HandshakeHash,
+    suite: &'static Tls13CipherSuite,
+    key_schedule: KeyScheduleTrafficWithClientFinishedPending,
+    send_tickets: usize,
+}
+
+impl State<ServerConnectionData> for ExpectCertificateOrCompressedCertificate {
+    fn handle<'m>(
+        self: Box<Self>,
+        cx: &mut ServerContext<'_>,
+        m: Message<'m>,
+    ) -> hs::NextStateOrError<'m>
+    where
+        Self: 'm,
+    {
+        match m.payload {
+            MessagePayload::Handshake {
+                parsed:
+                    HandshakeMessagePayload {
+                        payload: HandshakePayload::CertificateTls13(..),
+                        ..
+                    },
+                ..
+            } => Box::new(ExpectCertificate {
+                config: self.config,
+                transcript: self.transcript,
+                suite: self.suite,
+                key_schedule: self.key_schedule,
+                send_tickets: self.send_tickets,
+                message_already_in_transcript: false,
+            })
+            .handle(cx, m),
+
+            MessagePayload::Handshake {
+                parsed:
+                    HandshakeMessagePayload {
+                        payload: HandshakePayload::CompressedCertificate(..),
+                        ..
+                    },
+                ..
+            } => Box::new(ExpectCompressedCertificate {
+                config: self.config,
+                transcript: self.transcript,
+                suite: self.suite,
+                key_schedule: self.key_schedule,
+                send_tickets: self.send_tickets,
+            })
+            .handle(cx, m),
+
+            payload => Err(inappropriate_handshake_message(
+                &payload,
+                &[ContentType::Handshake],
+                &[
+                    HandshakeType::Certificate,
+                    HandshakeType::CompressedCertificate,
+                ],
+            )),
+        }
+    }
+
+    fn into_owned(self: Box<Self>) -> hs::NextState<'static> {
+        self
+    }
+}
+
+struct ExpectCompressedCertificate {
+    config: Arc<ServerConfig>,
+    transcript: HandshakeHash,
+    suite: &'static Tls13CipherSuite,
+    key_schedule: KeyScheduleTrafficWithClientFinishedPending,
+    send_tickets: usize,
+}
+
+impl State<ServerConnectionData> for ExpectCompressedCertificate {
+    fn handle<'m>(
+        mut self: Box<Self>,
+        cx: &mut ServerContext<'_>,
+        m: Message<'m>,
+    ) -> hs::NextStateOrError<'m>
+    where
+        Self: 'm,
+    {
+        self.transcript.add_message(&m);
+        let compressed_cert = require_handshake_msg_move!(
+            m,
+            HandshakeType::CompressedCertificate,
+            HandshakePayload::CompressedCertificate
+        )?;
+
+        let decompressor = match self
+            .config
+            .cert_decompressors
+            .iter()
+            .find(|item| item.algorithm() == compressed_cert.alg)
+        {
+            Some(dec) => dec,
+            None => {
+                return Err(cx.common.send_fatal_alert(
+                    AlertDescription::BadCertificate,
+                    PeerMisbehaved::SelectedUnofferedCertCompression,
+                ));
+            }
+        };
+
+        if compressed_cert.uncompressed_len as usize > CERTIFICATE_MAX_SIZE_LIMIT {
+            return Err(cx.common.send_fatal_alert(
+                AlertDescription::BadCertificate,
+                InvalidMessage::MessageTooLarge,
+            ));
+        }
+
+        let mut decompress_buffer = vec![0u8; compressed_cert.uncompressed_len as usize];
+        if let Err(compress::DecompressionFailed) =
+            decompressor.decompress(compressed_cert.compressed.0.bytes(), &mut decompress_buffer)
+        {
+            return Err(cx.common.send_fatal_alert(
+                AlertDescription::BadCertificate,
+                PeerMisbehaved::InvalidCertCompression,
+            ));
+        }
+
+        let cert_payload =
+            match CertificatePayloadTls13::read(&mut Reader::init(&decompress_buffer)) {
+                Ok(cm) => cm,
+                Err(err) => {
+                    return Err(cx
+                        .common
+                        .send_fatal_alert(AlertDescription::BadCertificate, err));
+                }
+            };
+        trace!(
+            "Client certificate decompressed using {:?} ({} bytes -> {})",
+            compressed_cert.alg,
+            compressed_cert
+                .compressed
+                .0
+                .bytes()
+                .len(),
+            compressed_cert.uncompressed_len,
+        );
+
+        let m = Message {
+            version: ProtocolVersion::TLSv1_3,
+            payload: MessagePayload::handshake(HandshakeMessagePayload {
+                typ: HandshakeType::Certificate,
+                payload: HandshakePayload::CertificateTls13(cert_payload.into_owned()),
+            }),
+        };
+
+        Box::new(ExpectCertificate {
+            config: self.config,
+            transcript: self.transcript,
+            suite: self.suite,
+            key_schedule: self.key_schedule,
+            send_tickets: self.send_tickets,
+            message_already_in_transcript: true,
+        })
+        .handle(cx, m)
+    }
+
+    fn into_owned(self: Box<Self>) -> hs::NextState<'static> {
+        self
+    }
+}
+
 struct ExpectCertificate {
     config: Arc<ServerConfig>,
     transcript: HandshakeHash,
     suite: &'static Tls13CipherSuite,
     key_schedule: KeyScheduleTrafficWithClientFinishedPending,
     send_tickets: usize,
+    message_already_in_transcript: bool,
 }
 
 impl State<ServerConnectionData> for ExpectCertificate {
@@ -850,7 +1093,9 @@ impl State<ServerConnectionData> for ExpectCertificate {
     where
         Self: 'm,
     {
-        self.transcript.add_message(&m);
+        if !self.message_already_in_transcript {
+            self.transcript.add_message(&m);
+        }
         let certp = require_handshake_msg_move!(
             m,
             HandshakeType::Certificate,
@@ -863,7 +1108,7 @@ impl State<ServerConnectionData> for ExpectCertificate {
             return Err(PeerMisbehaved::UnsolicitedCertExtension.into());
         }
 
-        let client_cert = certp.convert();
+        let client_cert = certp.into_certificate_chain();
 
         let mandatory = self
             .config
@@ -907,7 +1152,7 @@ impl State<ServerConnectionData> for ExpectCertificate {
             suite: self.suite,
             transcript: self.transcript,
             key_schedule: self.key_schedule,
-            client_cert,
+            client_cert: client_cert.into_owned(),
             send_tickets: self.send_tickets,
         }))
     }
