@@ -1,14 +1,13 @@
+use alloc::boxed::Box;
+use alloc::string::ToString;
+
 use crate::common_state::{CommonState, Side};
 use crate::crypto::cipher::{AeadKey, Iv, MessageDecrypter};
 use crate::crypto::tls13::{expand, Hkdf, HkdfExpander, OkmBlock, OutputLengthError};
-use crate::crypto::{hash, hmac, ActiveKeyExchange};
+use crate::crypto::{hash, hmac, SharedSecret};
 use crate::error::Error;
-use crate::quic;
 use crate::suites::PartiallyExtractedSecrets;
-use crate::{KeyLog, Tls13CipherSuite};
-
-use alloc::boxed::Box;
-use alloc::string::ToString;
+use crate::{quic, KeyLog, Tls13CipherSuite};
 
 /// Key schedule maintenance for TLS1.3
 
@@ -24,6 +23,8 @@ enum SecretKind {
     ExporterMasterSecret,
     ResumptionMasterSecret,
     DerivedSecret,
+    ServerEchConfirmationSecret,
+    ServerEchHrrConfirmationSecret,
 }
 
 impl SecretKind {
@@ -39,6 +40,10 @@ impl SecretKind {
             ExporterMasterSecret => b"exp master",
             ResumptionMasterSecret => b"res master",
             DerivedSecret => b"derived",
+            // https://datatracker.ietf.org/doc/html/draft-ietf-tls-esni-18#section-7.2
+            ServerEchConfirmationSecret => b"ech accept confirmation",
+            // https://datatracker.ietf.org/doc/html/draft-ietf-tls-esni-18#section-7.2.1
+            ServerEchHrrConfirmationSecret => b"hrr ech accept confirmation",
         }
     }
 
@@ -143,12 +148,11 @@ impl KeySchedulePreHandshake {
 
     pub(crate) fn into_handshake(
         mut self,
-        kx: Box<dyn ActiveKeyExchange>,
-        peer_public_key: &[u8],
-    ) -> Result<KeyScheduleHandshakeStart, Error> {
+        shared_secret: SharedSecret,
+    ) -> KeyScheduleHandshakeStart {
         self.ks
-            .input_from_key_exchange(kx, peer_public_key)?;
-        Ok(KeyScheduleHandshakeStart { ks: self.ks })
+            .input_secret(shared_secret.secret_bytes());
+        KeyScheduleHandshakeStart { ks: self.ks }
     }
 }
 
@@ -209,6 +213,30 @@ impl KeyScheduleHandshakeStart {
         new
     }
 
+    pub(crate) fn server_ech_confirmation_secret(
+        &mut self,
+        client_hello_inner_random: &[u8],
+        hs_hash: hash::Output,
+    ) -> [u8; 8] {
+        /*
+        Per ietf-tls-esni-17 section 7.2:
+        <https://datatracker.ietf.org/doc/html/draft-ietf-tls-esni-17#section-7.2>
+        accept_confirmation = HKDF-Expand-Label(
+          HKDF-Extract(0, ClientHelloInner.random),
+          "ech accept confirmation",
+          transcript_ech_conf,8)
+         */
+        hkdf_expand_label(
+            self.ks
+                .suite
+                .hkdf_provider
+                .extract_from_secret(None, client_hello_inner_random)
+                .as_ref(),
+            SecretKind::ServerEchConfirmationSecret.to_bytes(),
+            hs_hash.as_ref(),
+        )
+    }
+
     fn into_handshake(
         self,
         hs_hash: hash::Output,
@@ -607,7 +635,6 @@ impl KeySchedule {
     }
 
     /// Input the given secret.
-    #[cfg(all(test, any(feature = "ring", feature = "aws_lc_rs")))]
     fn input_secret(&mut self, secret: &[u8]) {
         let salt = self.derive_for_empty_hash(SecretKind::DerivedSecret);
         self.current = self
@@ -616,20 +643,6 @@ impl KeySchedule {
             .extract_from_secret(Some(salt.as_ref()), secret);
     }
 
-    /// Input the shared secret resulting from completing the given key exchange.
-    fn input_from_key_exchange(
-        &mut self,
-        kx: Box<dyn ActiveKeyExchange>,
-        peer_public_key: &[u8],
-    ) -> Result<(), Error> {
-        let salt = self.derive_for_empty_hash(SecretKind::DerivedSecret);
-        self.current = self
-            .suite
-            .hkdf_provider
-            .extract_from_kx_shared_secret(Some(salt.as_ref()), kx, peer_public_key)?;
-        Ok(())
-    }
-
     /// Derive a secret of given `kind`, using current handshake hash `hs_hash`.
     fn derive(&self, kind: SecretKind, hs_hash: &[u8]) -> OkmBlock {
         hkdf_expand_label_block(self.current.as_ref(), kind.to_bytes(), hs_hash)
@@ -792,6 +805,29 @@ fn hkdf_expand_label_slice(
     })
 }
 
+pub(crate) fn server_ech_hrr_confirmation_secret(
+    hkdf_provider: &'static dyn Hkdf,
+    client_hello_inner_random: &[u8],
+    hs_hash: hash::Output,
+) -> [u8; 8] {
+    /*
+    Per ietf-tls-esni-17 section 7.2.1:
+    <https://datatracker.ietf.org/doc/html/draft-ietf-tls-esni-17#section-7.2.1>
+    hrr_accept_confirmation = HKDF-Expand-Label(
+      HKDF-Extract(0, ClientHelloInner1.random),
+      "hrr ech accept confirmation",
+      transcript_hrr_ech_conf,
+      8)
+     */
+    hkdf_expand_label(
+        hkdf_provider
+            .extract_from_secret(None, client_hello_inner_random)
+            .as_ref(),
+        SecretKind::ServerEchHrrConfirmationSecret.to_bytes(),
+        hs_hash.as_ref(),
+    )
+}
+
 pub(crate) fn derive_traffic_key(expander: &dyn HkdfExpander, aead_key_len: usize) -> AeadKey {
     hkdf_expand_label_aead_key(expander, aead_key_len, b"key", &[])
 }
@@ -828,13 +864,14 @@ where
     f(expander, info)
 }
 
-#[cfg(all(test, any(feature = "ring", feature = "aws_lc_rs")))]
-mod tests {
+test_for_each_provider! {
     use core::fmt::Debug;
+    use std::vec;
+    use std::prelude::v1::*;
 
     use super::{derive_traffic_iv, derive_traffic_key, KeySchedule, SecretKind};
-    use crate::test_provider::ring_like::aead;
-    use crate::test_provider::tls13::{
+    use provider::ring_like::aead;
+    use provider::tls13::{
         TLS13_AES_128_GCM_SHA256_INTERNAL, TLS13_CHACHA20_POLY1305_SHA256_INTERNAL,
     };
     use crate::KeyLog;
@@ -1010,15 +1047,13 @@ mod tests {
     }
 }
 
-#[cfg(bench)]
-mod benchmarks {
-    #[cfg(any(feature = "ring", feature = "aws_lc_rs"))]
+bench_for_each_provider! {
     #[bench]
     fn bench_sha256(b: &mut test::Bencher) {
         use core::fmt::Debug;
 
         use super::{derive_traffic_iv, derive_traffic_key, KeySchedule, SecretKind};
-        use crate::test_provider::tls13::TLS13_CHACHA20_POLY1305_SHA256_INTERNAL;
+        use provider::tls13::TLS13_CHACHA20_POLY1305_SHA256_INTERNAL;
         use crate::KeyLog;
 
         fn extract_traffic_secret(ks: &KeySchedule, kind: SecretKind) {
