@@ -10,14 +10,15 @@ use pki_types::{
 };
 use rustls::client::danger::{HandshakeSignatureValid, ServerCertVerified, ServerCertVerifier};
 use rustls::client::{ServerCertVerifierBuilder, WebPkiServerVerifier};
+use rustls::crypto::cipher::{InboundOpaqueMessage, MessageDecrypter, MessageEncrypter};
 use rustls::crypto::CryptoProvider;
-use rustls::internal::msgs::codec::Reader;
+use rustls::internal::msgs::codec::{Codec, Reader};
 use rustls::internal::msgs::message::{Message, OutboundOpaqueMessage, PlainMessage};
 use rustls::server::{ClientCertVerifierBuilder, WebPkiClientVerifier};
 use rustls::{
-    ClientConfig, ClientConnection, Connection, ConnectionCommon, DigitallySignedStruct, Error,
-    ProtocolVersion, RootCertStore, ServerConfig, ServerConnection, SideData, SignatureScheme,
-    SupportedCipherSuite,
+    ClientConfig, ClientConnection, Connection, ConnectionCommon, ContentType,
+    DigitallySignedStruct, Error, ProtocolVersion, RootCertStore, ServerConfig, ServerConnection,
+    SideData, SignatureScheme, SupportedCipherSuite,
 };
 use webpki::anchor_from_trusted_cert;
 
@@ -53,18 +54,16 @@ embed_files! {
     (ECDSA_P256_CLIENT_CHAIN, "ecdsa-p256", "client.chain");
     (ECDSA_P256_CLIENT_FULLCHAIN, "ecdsa-p256", "client.fullchain");
     (ECDSA_P256_CLIENT_KEY, "ecdsa-p256", "client.key");
-    (ECDSA_P256_CLIENT_REQ, "ecdsa-p256", "client.req");
     (ECDSA_P256_END_CRL_PEM, "ecdsa-p256", "end.revoked.crl.pem");
     (ECDSA_P256_CLIENT_CRL_PEM, "ecdsa-p256", "client.revoked.crl.pem");
     (ECDSA_P256_INTERMEDIATE_CRL_PEM, "ecdsa-p256", "inter.revoked.crl.pem");
+    (ECDSA_P256_EXPIRED_CRL_PEM, "ecdsa-p256", "end.expired.crl.pem");
     (ECDSA_P256_END_CERT, "ecdsa-p256", "end.cert");
     (ECDSA_P256_END_CHAIN, "ecdsa-p256", "end.chain");
     (ECDSA_P256_END_FULLCHAIN, "ecdsa-p256", "end.fullchain");
     (ECDSA_P256_END_KEY, "ecdsa-p256", "end.key");
-    (ECDSA_P256_END_REQ, "ecdsa-p256", "end.req");
     (ECDSA_P256_INTER_CERT, "ecdsa-p256", "inter.cert");
     (ECDSA_P256_INTER_KEY, "ecdsa-p256", "inter.key");
-    (ECDSA_P256_INTER_REQ, "ecdsa-p256", "inter.req");
 
     (ECDSA_P384_CA_CERT, "ecdsa-p384", "ca.cert");
     (ECDSA_P384_CA_DER, "ecdsa-p384", "ca.der");
@@ -73,18 +72,16 @@ embed_files! {
     (ECDSA_P384_CLIENT_CHAIN, "ecdsa-p384", "client.chain");
     (ECDSA_P384_CLIENT_FULLCHAIN, "ecdsa-p384", "client.fullchain");
     (ECDSA_P384_CLIENT_KEY, "ecdsa-p384", "client.key");
-    (ECDSA_P384_CLIENT_REQ, "ecdsa-p384", "client.req");
     (ECDSA_P384_END_CRL_PEM, "ecdsa-p384", "end.revoked.crl.pem");
     (ECDSA_P384_CLIENT_CRL_PEM, "ecdsa-p384", "client.revoked.crl.pem");
     (ECDSA_P384_INTERMEDIATE_CRL_PEM, "ecdsa-p384", "inter.revoked.crl.pem");
+    (ECDSA_P384_EXPIRED_CRL_PEM, "ecdsa-p384", "end.expired.crl.pem");
     (ECDSA_P384_END_CERT, "ecdsa-p384", "end.cert");
     (ECDSA_P384_END_CHAIN, "ecdsa-p384", "end.chain");
     (ECDSA_P384_END_FULLCHAIN, "ecdsa-p384", "end.fullchain");
     (ECDSA_P384_END_KEY, "ecdsa-p384", "end.key");
-    (ECDSA_P384_END_REQ, "ecdsa-p384", "end.req");
     (ECDSA_P384_INTER_CERT, "ecdsa-p384", "inter.cert");
     (ECDSA_P384_INTER_KEY, "ecdsa-p384", "inter.key");
-    (ECDSA_P384_INTER_REQ, "ecdsa-p384", "inter.req");
 
     (ECDSA_P521_CA_CERT, "ecdsa-p521", "ca.cert");
     (ECDSA_P521_CA_DER, "ecdsa-p521", "ca.der");
@@ -93,18 +90,16 @@ embed_files! {
     (ECDSA_P521_CLIENT_CHAIN, "ecdsa-p521", "client.chain");
     (ECDSA_P521_CLIENT_FULLCHAIN, "ecdsa-p521", "client.fullchain");
     (ECDSA_P521_CLIENT_KEY, "ecdsa-p521", "client.key");
-    (ECDSA_P521_CLIENT_REQ, "ecdsa-p521", "client.req");
     (ECDSA_P521_END_CRL_PEM, "ecdsa-p521", "end.revoked.crl.pem");
     (ECDSA_P521_CLIENT_CRL_PEM, "ecdsa-p521", "client.revoked.crl.pem");
     (ECDSA_P521_INTERMEDIATE_CRL_PEM, "ecdsa-p521", "inter.revoked.crl.pem");
+    (ECDSA_P521_EXPIRED_CRL_PEM, "ecdsa-p521", "end.expired.crl.pem");
     (ECDSA_P521_END_CERT, "ecdsa-p521", "end.cert");
     (ECDSA_P521_END_CHAIN, "ecdsa-p521", "end.chain");
     (ECDSA_P521_END_FULLCHAIN, "ecdsa-p521", "end.fullchain");
     (ECDSA_P521_END_KEY, "ecdsa-p521", "end.key");
-    (ECDSA_P521_END_REQ, "ecdsa-p521", "end.req");
     (ECDSA_P521_INTER_CERT, "ecdsa-p521", "inter.cert");
     (ECDSA_P521_INTER_KEY, "ecdsa-p521", "inter.key");
-    (ECDSA_P521_INTER_REQ, "ecdsa-p521", "inter.req");
 
     (EDDSA_CA_CERT, "eddsa", "ca.cert");
     (EDDSA_CA_DER, "eddsa", "ca.der");
@@ -113,40 +108,70 @@ embed_files! {
     (EDDSA_CLIENT_CHAIN, "eddsa", "client.chain");
     (EDDSA_CLIENT_FULLCHAIN, "eddsa", "client.fullchain");
     (EDDSA_CLIENT_KEY, "eddsa", "client.key");
-    (EDDSA_CLIENT_REQ, "eddsa", "client.req");
     (EDDSA_END_CRL_PEM, "eddsa", "end.revoked.crl.pem");
     (EDDSA_CLIENT_CRL_PEM, "eddsa", "client.revoked.crl.pem");
     (EDDSA_INTERMEDIATE_CRL_PEM, "eddsa", "inter.revoked.crl.pem");
+    (EDDSA_EXPIRED_CRL_PEM, "eddsa", "end.expired.crl.pem");
     (EDDSA_END_CERT, "eddsa", "end.cert");
     (EDDSA_END_CHAIN, "eddsa", "end.chain");
     (EDDSA_END_FULLCHAIN, "eddsa", "end.fullchain");
     (EDDSA_END_KEY, "eddsa", "end.key");
-    (EDDSA_END_REQ, "eddsa", "end.req");
     (EDDSA_INTER_CERT, "eddsa", "inter.cert");
     (EDDSA_INTER_KEY, "eddsa", "inter.key");
-    (EDDSA_INTER_REQ, "eddsa", "inter.req");
 
-    (RSA_CA_CERT, "rsa", "ca.cert");
-    (RSA_CA_DER, "rsa", "ca.der");
-    (RSA_CA_KEY, "rsa", "ca.key");
-    (RSA_CLIENT_CERT, "rsa", "client.cert");
-    (RSA_CLIENT_CHAIN, "rsa", "client.chain");
-    (RSA_CLIENT_FULLCHAIN, "rsa", "client.fullchain");
-    (RSA_CLIENT_KEY, "rsa", "client.key");
-    (RSA_CLIENT_REQ, "rsa", "client.req");
-    (RSA_CLIENT_RSA, "rsa", "client.rsa");
-    (RSA_END_CRL_PEM, "rsa", "end.revoked.crl.pem");
-    (RSA_CLIENT_CRL_PEM, "rsa", "client.revoked.crl.pem");
-    (RSA_INTERMEDIATE_CRL_PEM, "rsa", "inter.revoked.crl.pem");
-    (RSA_END_CERT, "rsa", "end.cert");
-    (RSA_END_CHAIN, "rsa", "end.chain");
-    (RSA_END_FULLCHAIN, "rsa", "end.fullchain");
-    (RSA_END_KEY, "rsa", "end.key");
-    (RSA_END_REQ, "rsa", "end.req");
-    (RSA_END_RSA, "rsa", "end.rsa");
-    (RSA_INTER_CERT, "rsa", "inter.cert");
-    (RSA_INTER_KEY, "rsa", "inter.key");
-    (RSA_INTER_REQ, "rsa", "inter.req");
+    (RSA_2048_CA_CERT, "rsa-2048", "ca.cert");
+    (RSA_2048_CA_DER, "rsa-2048", "ca.der");
+    (RSA_2048_CA_KEY, "rsa-2048", "ca.key");
+    (RSA_2048_CLIENT_CERT, "rsa-2048", "client.cert");
+    (RSA_2048_CLIENT_CHAIN, "rsa-2048", "client.chain");
+    (RSA_2048_CLIENT_FULLCHAIN, "rsa-2048", "client.fullchain");
+    (RSA_2048_CLIENT_KEY, "rsa-2048", "client.key");
+    (RSA_2048_END_CRL_PEM, "rsa-2048", "end.revoked.crl.pem");
+    (RSA_2048_CLIENT_CRL_PEM, "rsa-2048", "client.revoked.crl.pem");
+    (RSA_2048_INTERMEDIATE_CRL_PEM, "rsa-2048", "inter.revoked.crl.pem");
+    (RSA_2048_EXPIRED_CRL_PEM, "rsa-2048", "end.expired.crl.pem");
+    (RSA_2048_END_CERT, "rsa-2048", "end.cert");
+    (RSA_2048_END_CHAIN, "rsa-2048", "end.chain");
+    (RSA_2048_END_FULLCHAIN, "rsa-2048", "end.fullchain");
+    (RSA_2048_END_KEY, "rsa-2048", "end.key");
+    (RSA_2048_INTER_CERT, "rsa-2048", "inter.cert");
+    (RSA_2048_INTER_KEY, "rsa-2048", "inter.key");
+
+    (RSA_3072_CA_CERT, "rsa-3072", "ca.cert");
+    (RSA_3072_CA_DER, "rsa-3072", "ca.der");
+    (RSA_3072_CA_KEY, "rsa-3072", "ca.key");
+    (RSA_3072_CLIENT_CERT, "rsa-3072", "client.cert");
+    (RSA_3072_CLIENT_CHAIN, "rsa-3072", "client.chain");
+    (RSA_3072_CLIENT_FULLCHAIN, "rsa-3072", "client.fullchain");
+    (RSA_3072_CLIENT_KEY, "rsa-3072", "client.key");
+    (RSA_3072_END_CRL_PEM, "rsa-3072", "end.revoked.crl.pem");
+    (RSA_3072_CLIENT_CRL_PEM, "rsa-3072", "client.revoked.crl.pem");
+    (RSA_3072_INTERMEDIATE_CRL_PEM, "rsa-3072", "inter.revoked.crl.pem");
+    (RSA_3072_EXPIRED_CRL_PEM, "rsa-3072", "end.expired.crl.pem");
+    (RSA_3072_END_CERT, "rsa-3072", "end.cert");
+    (RSA_3072_END_CHAIN, "rsa-3072", "end.chain");
+    (RSA_3072_END_FULLCHAIN, "rsa-3072", "end.fullchain");
+    (RSA_3072_END_KEY, "rsa-3072", "end.key");
+    (RSA_3072_INTER_CERT, "rsa-3072", "inter.cert");
+    (RSA_3072_INTER_KEY, "rsa-3072", "inter.key");
+
+    (RSA_4096_CA_CERT, "rsa-4096", "ca.cert");
+    (RSA_4096_CA_DER, "rsa-4096", "ca.der");
+    (RSA_4096_CA_KEY, "rsa-4096", "ca.key");
+    (RSA_4096_CLIENT_CERT, "rsa-4096", "client.cert");
+    (RSA_4096_CLIENT_CHAIN, "rsa-4096", "client.chain");
+    (RSA_4096_CLIENT_FULLCHAIN, "rsa-4096", "client.fullchain");
+    (RSA_4096_CLIENT_KEY, "rsa-4096", "client.key");
+    (RSA_4096_END_CRL_PEM, "rsa-4096", "end.revoked.crl.pem");
+    (RSA_4096_CLIENT_CRL_PEM, "rsa-4096", "client.revoked.crl.pem");
+    (RSA_4096_INTERMEDIATE_CRL_PEM, "rsa-4096", "inter.revoked.crl.pem");
+    (RSA_4096_EXPIRED_CRL_PEM, "rsa-4096", "end.expired.crl.pem");
+    (RSA_4096_END_CERT, "rsa-4096", "end.cert");
+    (RSA_4096_END_CHAIN, "rsa-4096", "end.chain");
+    (RSA_4096_END_FULLCHAIN, "rsa-4096", "end.fullchain");
+    (RSA_4096_END_KEY, "rsa-4096", "end.key");
+    (RSA_4096_INTER_CERT, "rsa-4096", "inter.cert");
+    (RSA_4096_INTER_KEY, "rsa-4096", "inter.key");
 }
 
 pub fn transfer(
@@ -243,7 +268,9 @@ where
 
 #[derive(Clone, Copy, Debug, PartialEq)]
 pub enum KeyType {
-    Rsa,
+    Rsa2048,
+    Rsa3072,
+    Rsa4096,
     EcdsaP256,
     EcdsaP384,
     EcdsaP521,
@@ -251,7 +278,9 @@ pub enum KeyType {
 }
 
 pub static ALL_KEY_TYPES: &[KeyType] = &[
-    KeyType::Rsa,
+    KeyType::Rsa2048,
+    KeyType::Rsa3072,
+    KeyType::Rsa4096,
     KeyType::EcdsaP256,
     KeyType::EcdsaP384,
     #[cfg(all(not(feature = "ring"), feature = "aws_lc_rs"))]
@@ -262,7 +291,9 @@ pub static ALL_KEY_TYPES: &[KeyType] = &[
 impl KeyType {
     fn bytes_for(&self, part: &str) -> &'static [u8] {
         match self {
-            Self::Rsa => bytes_for("rsa", part),
+            Self::Rsa2048 => bytes_for("rsa-2048", part),
+            Self::Rsa3072 => bytes_for("rsa-3072", part),
+            Self::Rsa4096 => bytes_for("rsa-4096", part),
             Self::EcdsaP256 => bytes_for("ecdsa-p256", part),
             Self::EcdsaP384 => bytes_for("ecdsa-p384", part),
             Self::EcdsaP521 => bytes_for("ecdsa-p521", part),
@@ -292,18 +323,22 @@ impl KeyType {
     }
 
     pub fn end_entity_crl(&self) -> CertificateRevocationListDer<'static> {
-        self.get_crl("end")
+        self.get_crl("end", "revoked")
     }
 
     pub fn client_crl(&self) -> CertificateRevocationListDer<'static> {
-        self.get_crl("client")
+        self.get_crl("client", "revoked")
     }
 
     pub fn intermediate_crl(&self) -> CertificateRevocationListDer<'static> {
-        self.get_crl("inter")
+        self.get_crl("inter", "revoked")
     }
 
-    fn get_client_key(&self) -> PrivateKeyDer<'static> {
+    pub fn end_entity_crl_expired(&self) -> CertificateRevocationListDer<'static> {
+        self.get_crl("end", "expired")
+    }
+
+    pub fn get_client_key(&self) -> PrivateKeyDer<'static> {
         PrivateKeyDer::Pkcs8(
             rustls_pemfile::pkcs8_private_keys(&mut io::BufReader::new(
                 self.bytes_for("client.key"),
@@ -314,9 +349,9 @@ impl KeyType {
         )
     }
 
-    fn get_crl(&self, role: &str) -> CertificateRevocationListDer<'static> {
+    fn get_crl(&self, role: &str, r#type: &str) -> CertificateRevocationListDer<'static> {
         rustls_pemfile::crls(&mut io::BufReader::new(
-            self.bytes_for(&format!("{role}.revoked.crl.pem")),
+            self.bytes_for(&format!("{role}.{type}.crl.pem")),
         ))
         .map(|result| result.unwrap())
         .next() // We only expect one CRL.
@@ -325,7 +360,15 @@ impl KeyType {
 
     pub fn ca_distinguished_name(&self) -> &'static [u8] {
         match self {
-            KeyType::Rsa => &b"0\x1a1\x180\x16\x06\x03U\x04\x03\x0c\x0fponytown RSA CA"[..],
+            KeyType::Rsa2048 => {
+                &b"0\x1f1\x1d0\x1b\x06\x03U\x04\x03\x0c\x14ponytown RSA 2048 CA"[..]
+            }
+            KeyType::Rsa3072 => {
+                &b"0\x1f1\x1d0\x1b\x06\x03U\x04\x03\x0c\x14ponytown RSA 3072 CA"[..]
+            }
+            KeyType::Rsa4096 => {
+                &b"0\x1f1\x1d0\x1b\x06\x03U\x04\x03\x0c\x14ponytown RSA 4096 CA"[..]
+            }
             KeyType::EcdsaP256 => {
                 &b"0\x211\x1f0\x1d\x06\x03U\x04\x03\x0c\x16ponytown ECDSA p256 CA"[..]
             }
@@ -747,6 +790,7 @@ pub struct MockServerVerifier {
     tls12_signature_error: Option<Error>,
     tls13_signature_error: Option<Error>,
     signature_schemes: Vec<SignatureScheme>,
+    expected_ocsp_response: Option<Vec<u8>>,
 }
 
 impl ServerCertVerifier for MockServerVerifier {
@@ -755,13 +799,16 @@ impl ServerCertVerifier for MockServerVerifier {
         end_entity: &CertificateDer<'_>,
         intermediates: &[CertificateDer<'_>],
         server_name: &ServerName<'_>,
-        oscp_response: &[u8],
+        ocsp_response: &[u8],
         now: UnixTime,
     ) -> Result<ServerCertVerified, Error> {
         println!(
             "verify_server_cert({:?}, {:?}, {:?}, {:?}, {:?})",
-            end_entity, intermediates, server_name, oscp_response, now
+            end_entity, intermediates, server_name, ocsp_response, now
         );
+        if let Some(expected_ocsp) = &self.expected_ocsp_response {
+            assert_eq!(expected_ocsp, ocsp_response);
+        }
         if let Some(error) = &self.cert_rejection_error {
             Err(error.clone())
         } else {
@@ -816,6 +863,13 @@ impl MockServerVerifier {
         }
     }
 
+    pub fn expects_ocsp_response(response: &[u8]) -> Self {
+        MockServerVerifier {
+            expected_ocsp_response: Some(response.to_vec()),
+            ..Default::default()
+        }
+    }
+
     pub fn rejects_certificate(err: Error) -> Self {
         MockServerVerifier {
             cert_rejection_error: Some(err),
@@ -859,6 +913,130 @@ impl Default for MockServerVerifier {
                 SignatureScheme::ECDSA_NISTP384_SHA384,
                 SignatureScheme::ECDSA_NISTP521_SHA512,
             ],
+            expected_ocsp_response: None,
         }
     }
 }
+
+/// This allows injection/receipt of raw messages into a post-handshake connection.
+///
+/// It consumes one of the peers, extracts its secrets, and then reconstitutes the
+/// message encrypter/decrypter.  It does not do fragmentation/joining.
+pub struct RawTls {
+    encrypter: Box<dyn MessageEncrypter>,
+    enc_seq: u64,
+    decrypter: Box<dyn MessageDecrypter>,
+    dec_seq: u64,
+}
+
+impl RawTls {
+    /// conn must be post-handshake, and must have been created with `enable_secret_extraction`
+    pub fn new_client(conn: ClientConnection) -> Self {
+        let suite = conn.negotiated_cipher_suite().unwrap();
+        Self::new(
+            suite,
+            conn.dangerous_extract_secrets()
+                .unwrap(),
+        )
+    }
+
+    /// conn must be post-handshake, and must have been created with `enable_secret_extraction`
+    pub fn new_server(conn: ServerConnection) -> Self {
+        let suite = conn.negotiated_cipher_suite().unwrap();
+        Self::new(
+            suite,
+            conn.dangerous_extract_secrets()
+                .unwrap(),
+        )
+    }
+
+    fn new(suite: SupportedCipherSuite, secrets: rustls::ExtractedSecrets) -> Self {
+        let rustls::ExtractedSecrets {
+            tx: (tx_seq, tx_keys),
+            rx: (rx_seq, rx_keys),
+        } = secrets;
+
+        let encrypter = match (tx_keys, suite) {
+            (
+                rustls::ConnectionTrafficSecrets::Aes256Gcm { key, iv },
+                SupportedCipherSuite::Tls13(tls13),
+            ) => tls13.aead_alg.encrypter(key, iv),
+
+            (
+                rustls::ConnectionTrafficSecrets::Aes256Gcm { key, iv },
+                SupportedCipherSuite::Tls12(tls12),
+            ) => tls12
+                .aead_alg
+                .encrypter(key, &iv.as_ref()[..4], &iv.as_ref()[4..]),
+
+            _ => todo!(),
+        };
+
+        let decrypter = match (rx_keys, suite) {
+            (
+                rustls::ConnectionTrafficSecrets::Aes256Gcm { key, iv },
+                SupportedCipherSuite::Tls13(tls13),
+            ) => tls13.aead_alg.decrypter(key, iv),
+
+            (
+                rustls::ConnectionTrafficSecrets::Aes256Gcm { key, iv },
+                SupportedCipherSuite::Tls12(tls12),
+            ) => tls12
+                .aead_alg
+                .decrypter(key, &iv.as_ref()[..4]),
+
+            _ => todo!(),
+        };
+
+        Self {
+            encrypter,
+            enc_seq: tx_seq,
+            decrypter,
+            dec_seq: rx_seq,
+        }
+    }
+
+    pub fn encrypt_and_send(
+        &mut self,
+        msg: &PlainMessage,
+        peer: &mut impl DerefMut<Target = ConnectionCommon<impl SideData>>,
+    ) {
+        let data = self
+            .encrypter
+            .encrypt(msg.borrow_outbound(), self.enc_seq)
+            .unwrap()
+            .encode();
+        self.enc_seq += 1;
+        peer.read_tls(&mut io::Cursor::new(data))
+            .unwrap();
+    }
+
+    pub fn receive_and_decrypt(
+        &mut self,
+        peer: &mut impl DerefMut<Target = ConnectionCommon<impl SideData>>,
+        f: impl Fn(Message),
+    ) {
+        let mut data = vec![];
+        peer.write_tls(&mut io::Cursor::new(&mut data))
+            .unwrap();
+
+        let mut reader = Reader::init(&data);
+        let content_type = ContentType::read(&mut reader).unwrap();
+        let version = ProtocolVersion::read(&mut reader).unwrap();
+        let len = u16::read(&mut reader).unwrap();
+        let left = &mut data[5..];
+        assert_eq!(len as usize, left.len());
+
+        let inbound = InboundOpaqueMessage::new(content_type, version, left);
+        let plain = self
+            .decrypter
+            .decrypt(inbound, self.dec_seq)
+            .unwrap();
+        self.dec_seq += 1;
+
+        let msg = Message::try_from(plain).unwrap();
+        println!("receive_and_decrypt: {msg:?}");
+
+        f(msg);
+    }
+}
